from blueman.main.PluginManager import StopException
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from gi.repository import Gtk
from gi.repository import GObject


class StatusIcon(AppletPlugin, Gtk.StatusIcon):
    __unloadable__ = False
    __icon__ = "blueman"

    def on_entry_changed(self, entry, ic, image):

        if ic.has_icon(self.get_option("icon")):
            icon = Gtk.STOCK_APPLY
        else:
            icon = Gtk.STOCK_CANCEL

        image.set_from_stock(icon, Gtk.IconSize.LARGE_TOOLBAR)

        if self.timeout:
            GObject.source_remove(self.timeout)

        self.timeout = GObject.timeout_add(1000, lambda: self.IconShouldChange())

    def widget_decorator(self, widget, name, options):
        entry = widget.get_children()[1]
        image = Gtk.Image()

        completion = Gtk.EntryCompletion()
        entry.set_completion(completion)

        liststore = Gtk.ListStore(GObject.TYPE_STRING)

        completion.set_model(liststore)

        completion.set_text_column(0)

        ic = Gtk.IconTheme.get_default()
        icons = ic.list_icons("Applications")
        for i in icons:
            liststore.append([i])

        if ic.has_icon(self.get_option("icon")):
            icon = Gtk.STOCK_APPLY
        else:
            icon = Gtk.STOCK_CANCEL

        image.set_from_stock(icon, Gtk.IconSize.LARGE_TOOLBAR)
        image.show()
        widget.pack_start(image, True, 0, 0)
        entry.connect("changed", self.on_entry_changed, ic, image)

    __options__ = {"icon": {"type": str,
                            "default": "blueman",
                            "name": _("Icon Name"),
                            "desc": _("Custom icon to use for the notification area"),
                            "decorator": widget_decorator
    }
    }

    FORCE_SHOW = 2
    SHOW = 1
    FORCE_HIDE = 0

    def on_load(self, applet):
        GObject.GObject.__init__(self)
        self.lines = {}
        self.pixbuf = None
        self.timeout = None

        #self.connect("size-changed", self.on_status_icon_resized)

        self.SetTextLine(0, _("Bluetooth Enabled"))

        AppletPlugin.add_method(self.on_query_status_icon_visibility)
        AppletPlugin.add_method(self.on_status_icon_query_icon)

        ic = Gtk.IconTheme.get_default()
        ic.connect("changed", self.on_icon_theme_changed)

        self.on_status_icon_resized()

    def on_icon_theme_changed(self, icon_theme):
        self.IconShouldChange()

    def on_power_state_changed(self, manager, state):
        if state:
            self.SetTextLine(0, _("Bluetooth Enabled"))
        else:
            self.SetTextLine(0, _("Bluetooth Disabled"))

        self.QueryVisibility()

    def QueryVisibility(self):

        rets = self.Applet.Plugins.Run("on_query_status_icon_visibility")
        if not StatusIcon.FORCE_HIDE in rets:
            if StatusIcon.FORCE_SHOW in rets:
                self.set_visible(True)
            else:
                if not self.Applet.Manager:
                    self.set_visible(False)
                    return

                try:
                    self.set_visible(self.Applet.Manager.list_adapters())
                except:
                    self.set_visible(False)
        else:
            self.set_visible(False)

    def set_visible(self, visible):
        self.props.visible = visible

    def SetTextLine(self, id, text):
        if text:
            self.lines[id] = text
        else:
            try:
                del self.lines[id]
            except:
                pass

        self.update_tooltip()


    def update_tooltip(self):
        s = ""
        keys = self.lines.keys()
        keys.sort()
        for k in keys:
            s += self.lines[k] + "\n"

        self.props.tooltip_markup = s[:-1]

    def IconShouldChange(self):
        self.on_status_icon_resized()

    def on_adapter_added(self, path):
        self.QueryVisibility()

    def on_adapter_removed(self, path):
        self.QueryVisibility()

    def on_manager_state_changed(self, state):
        self.QueryVisibility()

    def on_status_icon_resized(self):
        self.icon = "blueman"

        #p = GdkPixbuf.Pixbuf(GdkPixbuf.Colorspace.RGB, True, 8, 22, 22)
        #p.fill(0)


        #self.pixbuf.copy_area(0, 0, self.pixbuf.props.width, self.pixbuf.props.height, p, 5, 0)

        #self.pixbuf = p
        ic = Gtk.IconTheme.get_default()

        def callback(inst, ret):
            if ret != None:
                for i in ret:
                    if ic.has_icon(i):
                        self.icon = i
                        raise StopException

        self.Applet.Plugins.RunEx("on_status_icon_query_icon", callback)
        self.props.icon_name = self.icon

        return True

    def on_query_status_icon_visibility(self):
        return StatusIcon.SHOW

    def on_status_icon_query_icon(self):
        return None




import dbus
from gi.repository import GObject
from blueman.main.HalManager import HalManager
from blueman.Functions import dprint


class WrongType(Exception):
    pass


class NotAKillSwitch(Exception):
    pass


class KillSwitch(dbus.proxies.Interface):
    class __Switch(dbus.proxies.Interface):
        def __init__(self, udi):
            bus = dbus.SystemBus()
            obj = bus.get_object('org.freedesktop.Hal', udi)
            dbus.proxies.Interface.__init__(self, obj, 'org.freedesktop.Hal.Device.KillSwitch')

    def __init__(self, udi):
        self.udi = udi
        bus = dbus.SystemBus()
        obj = bus.get_object('org.freedesktop.Hal', udi)
        dbus.proxies.Interface.__init__(self, obj, 'org.freedesktop.Hal.Device')
        if self.QueryCapability("killswitch"):
            t = self.GetPropertyString("killswitch.type")
            if t != "bluetooth":
                raise WrongType
        else:
            raise NotAKillSwitch

        self.__switch = KillSwitch.__Switch(udi)

        self.hard = 0
        self.idx = self.udi
        self.type = 2  # RfkillType.BLUETOOTH

    @property
    def soft(self):
        try:
            return not self.GetPower()
        except:
            return False

    def SetPower(self, state):
        try:
            self.__switch.SetPower(state)
        except dbus.DBusException:
            dprint("Failed to toggle killswitch")

    def GetPower(self):
        return self.__switch.GetPower()


class Manager(GObject.GObject):
    __inst = None
    __gsignals__ = {
    'switch-changed': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
    'switch-added': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
    'switch-removed': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __new__(cls):
        if not Manager.__inst:
            return super(Manager, cls).__new__(cls)

        return Manager.__inst

    def __init__(self):
        if not Manager.__inst:
            GObject.GObject.__init__(self)
            Manager.__inst = self
            self.devices = []
            self.state = True
            self.HardBlocked = False
            dbus.SystemBus().watch_name_owner("org.freedesktop.Hal", self.hal_name_owner_changed)

    def hal_name_owner_changed(self, owner):
        for dev in self.devices:
            self.emit("switch-removed", dev)
        self.devices = []
        if owner != "":
            self.Hal = HalManager()
            self.__enumerate()
        else:
            self.Hal = None

    def __enumerate(self):
        self.state = True

        devs = self.Hal.FindDeviceByCapability("killswitch")
        for dev in devs:
            try:
                sw = KillSwitch(dev)
                self.devices.append(sw)
                self.state &= sw.GetPower()
                self.emit("switch-added", sw)
            except WrongType:
                pass

    def SetGlobalState(self, state, **kwargs):
        dprint("Setting killswitches to", state)

        for dev in self.devices:
            print("Setting", dev.udi, "to", state)
            dev.SetPower(state)
        if len(self.devices) == 0:
            self.state = True
        else:
            self.state = state

        if "reply_handler" in kwargs:
            kwargs["reply_handler"]()

    def GetGlobalState(self):
        try:
            self.state &= self.devices[0]
        except:
            return self.state
        else:
            return self.state





#!/usr/bin/python

cdef extern from "malloc.h":
	cdef void free(void *ptr)
	
cdef extern from "string.h":
	cdef char* strerror(int errnum)

cdef extern from "bluetooth/bluetooth.h":
	ctypedef struct bdaddr_t:
		unsigned char b[6]

	int ba2str(bdaddr_t *ba, char *str)
	int str2ba(char *str, bdaddr_t *ba)
	
cdef extern from "bluetooth/hci.h":	
	cdef struct hci_dev_stats:
		unsigned long int err_rx
		unsigned long int err_tx
		unsigned long int cmd_tx
		unsigned long int evt_rx
		unsigned long int acl_tx
		unsigned long int acl_rx
		unsigned long int sco_tx
		unsigned long int sco_rx
		unsigned long int byte_rx
		unsigned long int byte_tx


	cdef struct hci_dev_info:
		unsigned int dev_id
		char name[8]

		bdaddr_t bdaddr

		unsigned long int flags
		unsigned char type

		unsigned char features[8]

		unsigned long int pkt_type
		unsigned long int link_policy
		unsigned long int link_mode

		unsigned int acl_mtu
		unsigned int acl_pkts
		unsigned int sco_mtu
		unsigned int sco_pkts

		hci_dev_stats stat



cdef extern from "bluetooth/hci_lib.h":
	cdef int hci_devinfo(int dev_id, hci_dev_info *di)


cdef extern from "bluetooth/rfcomm.h":

	cdef struct rfcomm_dev_info:
		short id
		unsigned int flags
		unsigned short state
		bdaddr_t src
		bdaddr_t dst
		unsigned char channel
		
	cdef struct rfcomm_dev_list_req:
		unsigned short dev_num
		rfcomm_dev_info dev_info[0]
	

cdef extern from "libblueman.h":
	cdef struct conn_info_handles:
		unsigned int handle
		int dd
	

	cdef int connection_init(int dev_id, char *addr, conn_info_handles *ci)
	cdef int connection_get_rssi(conn_info_handles *ci, char *ret_rssi)
	cdef int connection_get_lq(conn_info_handles *ci, unsigned char *ret_lq)
	cdef int connection_get_tpl(conn_info_handles *ci, char *ret_tpl, unsigned char type)
	cdef int connection_close(conn_info_handles *ci)
	cdef int get_rfcomm_list(rfcomm_dev_list_req **ret)
	cdef float get_page_timeout(int hdev)
	cdef int _create_bridge(char* name)
	cdef int _destroy_bridge(char* name)
	cdef char* c_get_net_address "get_net_address" (char* iface, int i)
	cdef char** c_get_interface_list "get_interface_list" ()
	
cdef extern from "linux/sockios.h":
	cdef int SIOCGIFADDR
	cdef int SIOCGIFNETMASK
	
def get_net_address(iface):
	cdef char* addr
	if iface != None:
		addr = c_get_net_address(iface, SIOCGIFADDR)
		if addr == NULL:
			return None
		else:
			return addr
			
def get_net_netmask(iface):
	cdef char* addr
	if iface != None:
		addr = c_get_net_address(iface, SIOCGIFNETMASK)
		if addr == NULL:
			return None
		else:
			return addr

def get_net_interfaces():
	cdef char** ifaces
	cdef int i
	i = 0
	
	ifaces = c_get_interface_list()
	if ifaces == NULL:
		return []
		
	ret = []
	while 1:
		if ifaces[i] == NULL:
			break
		else:
			ret.append(ifaces[i])
			free(ifaces[i])
		i = i + 1
		
	free(ifaces)
	return ret

ERR = {
	-1:"Can't allocate memory",
	-2:"HCI device open failed",
	-3:"Not connected",
	-4:"Get connection info failed",
	-5:"Read RSSI failed",
	-6:"Read transmit power level request failed",
	-7:"Read Link quality failed",
	-8:"Getting rfcomm list failed",
	-9:"ERR_SOCKET_FAILED",
	-10:"ERR_CANT_READ_PAGE_TIMEOUT",
	-11:"ERR_READ_PAGE_TIMEOUT"
	}
	
RFCOMM_STATES = [
	"unknown",
	"connected",
	"clean",
	"bound",
	"listening",
	"connecting",
	"connecting",
	"config",
	"disconnecting",
	"closed"
]
RFCOMM_REUSE_DLC =	0
RFCOMM_RELEASE_ONHUP =	1
RFCOMM_HANGUP_NOW =	2
RFCOMM_TTY_ATTACHED =	3

def rfcomm_list():
	cdef rfcomm_dev_list_req *dl

	cdef char src[18]
	cdef char dst[18]

	
	res = get_rfcomm_list(&dl)
	if res < 0:
		raise Exception, ERR[res]
	
	devs = []
	for 0 <= i < dl.dev_num:
		ba2str(&dl.dev_info[i].src, src)
		ba2str(&dl.dev_info[i].dst, dst)
		
		devs.append(
			{	"id": dl.dev_info[i].id,
				"channel": dl.dev_info[i].channel,
				"flags": dl.dev_info[i].flags,
				"state": RFCOMM_STATES[dl.dev_info[i].state],
				"src": src,
				"dst": dst	
			})

	free(dl)

	return devs
	
import exceptions
class BridgeException(exceptions.Exception):
	def __init__(self, errno):
		self.errno = errno
		
	def __str__(self):
		return strerror(self.errno)
		


def create_bridge(name="pan1"):
	err = _create_bridge(name)
	if err < 0:
		raise BridgeException(-err)
	
def destroy_bridge(name="pan1"):
	err = _destroy_bridge(name)
	if err < 0:
		raise BridgeException(-err)


cdef class conn_info:
	cdef conn_info_handles ci
	cdef int hci
	
	def __init__(self, addr, hci_name="hci0"):
		
				
		self.hci = int(hci_name[3:])
		res = connection_init(self.hci, addr, &self.ci)
		if res < 0:
			raise Exception, ERR[res]
			
	def deinit(self):
		connection_close(&self.ci)
			
	def get_rssi(self):
		cdef char rssi
		res = connection_get_rssi(&self.ci, &rssi)
		if res < 0:
			raise Exception, ERR[res]
		
		return rssi
		
	def get_lq(self):
		cdef unsigned char lq
		res = connection_get_lq(&self.ci, &lq)
		if res < 0:
			raise Exception, ERR[res]
		
		return lq
		
	def get_tpl(self, tp=0):
		cdef char tpl
		res = connection_get_tpl(&self.ci, &tpl, tp)
		if res < 0:
			raise Exception, ERR[res]
		
		return tpl			

def page_timeout(hci_name="hci0"):
	dev_id = int(hci_name[3:])
	ret = get_page_timeout(dev_id)
	if ret < 0:
		raise Exception, ERR[ret]
	else:
		return ret
		
def device_info(hci_name="hci0"):
	cdef hci_dev_info di
	cdef int ret
	
	dev_id = int(hci_name[3:])
	
	res = hci_devinfo(dev_id, &di)
	
	cdef char addr[32]
	ba2str(&di.bdaddr, addr)
	
	feats = []
	for 0 <= i < 8: 
		feats.append(di.features[i])
	
	x = [("err_rx", di.stat.err_rx),
	("err_tx",di.stat.err_tx),
	("cmd_tx",di.stat.cmd_tx),
	("evt_rx",di.stat.evt_rx),
	("acl_tx",di.stat.acl_tx),
	("acl_rx",di.stat.acl_rx),
	("sco_tx",di.stat.sco_tx),
	("sco_rx",di.stat.sco_rx),
	("byte_rx",di.stat.byte_rx),
	("byte_tx",di.stat.byte_tx)]
	
	z = [("dev_id", di.dev_id),
	("name", di.name),
	("bdaddr",addr),
	("flags",di.flags),
	("type",di.type),
	("features",feats),
	("pkt_type",di.pkt_type),
	("link_policy",di.link_policy),
	("link_mode",di.link_mode),
	("acl_mtu",di.acl_mtu),
	("acl_pkts",di.acl_pkts),
	("sco_mtu",di.sco_pkts),
	("stat", dict(x))]

	return dict(z)

cdef extern from "X11/X.h":
	ctypedef unsigned long Time
	ctypedef struct Display

cdef extern from "libsn/sn-common.h":
	ctypedef struct SnDisplay
	ctypedef void (* SnDisplayErrorTrapPush) (SnDisplay *display, Display *xdisplay)
	ctypedef void (* SnDisplayErrorTrapPop)  (SnDisplay *display, Display *xdisplay)
	SnDisplay* sn_display_new (Display                *xdisplay,
                                       SnDisplayErrorTrapPush  push_trap_func,
                                       SnDisplayErrorTrapPop   pop_trap_func)
	cdef void       sn_display_ref             (SnDisplay              *display)
	cdef void       sn_display_unref           (SnDisplay              *display)

cdef extern from "libsn/sn-launcher.h":
	cdef struct SnLauncherContext
	cdef SnLauncherContext* sn_launcher_context_new (SnDisplay *display, int screen)
	cdef void        sn_launcher_context_ref               (SnLauncherContext *context)
	cdef void        sn_launcher_context_unref             (SnLauncherContext *context)


	cdef void        sn_launcher_context_initiate          (SnLauncherContext *context, char *launcher_name, char *launchee_name, Time timestamp)
	cdef void        sn_launcher_context_complete          (SnLauncherContext *context)
	cdef char*       sn_launcher_context_get_startup_id    (SnLauncherContext *context)
	cdef int         sn_launcher_context_get_initiated     (SnLauncherContext *context)

	cdef void        sn_launcher_context_setup_child_process (SnLauncherContext *context)

	cdef void sn_launcher_context_set_name        (SnLauncherContext *context,
		                                   char        *name)
	cdef void sn_launcher_context_set_description (SnLauncherContext *context,
		                                   char        *description)
	cdef void sn_launcher_context_set_workspace   (SnLauncherContext *context,
		                                  int                workspace)
	cdef void sn_launcher_context_set_wmclass     (SnLauncherContext *context,
		                                   char        *klass)
	cdef void sn_launcher_context_set_binary_name (SnLauncherContext *context,
		                                   char        *name)
	cdef void sn_launcher_context_set_icon_name   (SnLauncherContext *context,
		                                   char        *name)

	cdef void sn_launcher_context_set_extra_property (SnLauncherContext *context,
		                                      char        *name,
		                                      char        *value)


	cdef void sn_launcher_context_get_initiated_time   (SnLauncherContext *context,
		                                       long              *tv_sec,
		                                       long              *tv_usec)
	cdef void sn_launcher_context_get_last_active_time (SnLauncherContext *context,
		                                       long              *tv_sec,
		                                       long              *tv_usec)



cdef extern SnLauncherContext* GetSnLauncherContext()

cdef extern from "Python.h":
	cdef struct PyObject

cdef extern from "stdio.h":
	cdef int printf(char* format, ...)
	
cdef extern from "glib-object.h":
	ctypedef struct GObject
	
cdef extern from "pygobject.h":
	cdef GObject* pygobject_get(object)
	
cdef extern from "gdk/gdkx.h":
	ctypedef struct GdkDisplay
	cdef void gdk_error_trap_push ()
	cdef void gdk_error_trap_pop ()
	cdef Display* gdk_x11_display_get_xdisplay(GdkDisplay*)

cdef void sn_error_trap_push(SnDisplay *display, Display *xdisplay):
	gdk_error_trap_push ()
	
cdef void sn_error_trap_pop(SnDisplay *display, Display *xdisplay):
	gdk_error_trap_pop ()


cdef class sn_launcher:
	cdef SnLauncherContext* ctx
	

	def __cinit__(self, display, int screen):
		#from gi.repository import GdkX11
		#if type(display) != GdkX11.X11Display:
		#	raise TypeError, "Display must be a Gdk.DisplayX11"
			
		cdef GObject* dpy
		cdef SnDisplay* sn_dpy
		
		dpy = pygobject_get(display)
		
		if dpy != NULL:
			sn_dpy =  sn_display_new(gdk_x11_display_get_xdisplay(<GdkDisplay*>dpy), sn_error_trap_push, sn_error_trap_pop)
			self.ctx = sn_launcher_context_new(sn_dpy, screen)
			sn_display_unref(sn_dpy)
		else:
			raise RuntimeError, "GdkDisplay is NULL"
		
		
	def __dealloc__(self):
		if self.ctx == NULL:
			raise RuntimeError, "SnLauncherContext is NULL"
		sn_launcher_context_unref(self.ctx)

		
	def initiate(self, char* launcher_name, char* launchee_name, Time timestamp):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_initiate(self.ctx, launcher_name, launchee_name, timestamp)
		sn_launcher_context_unref(self.ctx)
	
	def complete(self):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_complete(self.ctx)
		sn_launcher_context_unref(self.ctx)
		
	def get_startup_id(self):
		cdef char* ret

		sn_launcher_context_ref(self.ctx)
		ret = sn_launcher_context_get_startup_id(self.ctx)
		sn_launcher_context_unref(self.ctx)
		if ret != NULL:
			return ret
		else:
			return None
		
	def get_initiated(self):
		sn_launcher_context_ref(self.ctx)
		ret = bool(sn_launcher_context_get_initiated(self.ctx))
		sn_launcher_context_unref(self.ctx)
		return ret
	
	def setup_child_process(self):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_setup_child_process(self.ctx)
		sn_launcher_context_unref(self.ctx)
	
	def set_name(self, char* name):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_name(self.ctx, name)
		sn_launcher_context_unref(self.ctx)
		
	def set_description(self, char* descr):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_description(self.ctx, descr)
		sn_launcher_context_unref(self.ctx)
		
	def set_workspace(self, int workspace):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_workspace(self.ctx, workspace)
		sn_launcher_context_unref(self.ctx)
		
	def set_wmclass(self, char* klass):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_wmclass(self.ctx, klass)
		sn_launcher_context_unref(self.ctx)
		
	def set_binary_name(self, char* name):
		sn_launcher_context_ref(self.ctx)
		
		sn_launcher_context_set_binary_name(self.ctx, name)
		sn_launcher_context_unref(self.ctx)
		
	def set_icon_name(self, char* name):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_icon_name(self.ctx, name)
		sn_launcher_context_unref(self.ctx)
		
	def set_extra_property(self, char* key, char* value):
		sn_launcher_context_ref(self.ctx)
		sn_launcher_context_set_extra_property(self.ctx, key, value)
		sn_launcher_context_unref(self.ctx)
		
	def get_initiated_time(self):
		sn_launcher_context_ref(self.ctx)
		cdef long tv_sec
		cdef long tv_usec
		sn_launcher_context_get_initiated_time(self.ctx, &tv_sec, &tv_usec)
		sn_launcher_context_unref(self.ctx)
		return (tv_sec, tv_usec)
		
	def get_last_active_time(self):
		sn_launcher_context_ref(self.ctx)
		cdef long tv_sec
		cdef long tv_usec
		sn_launcher_context_get_last_active_time(self.ctx, &tv_sec, &tv_usec)
		sn_launcher_context_unref(self.ctx)
		return (tv_sec, tv_usec)


cdef extern from "modem-prober.h":
	cdef void c_probe_modem "probe_modem" (char* device, object callback)
	cdef void c_set_probe_debug "set_probe_debug" (int debug)

def probe_modem(node, callback):

	if not callable(callback):
		raise TypeError, "callback must be callable"
		
	if node != None:
		c_probe_modem(node, callback)
	else:
		raise TypeError, "device node must not be None"
		
def set_probe_debug(enable):
	c_set_probe_debug(int(enable))

cdef extern from "glib.h":
	cdef char* g_get_user_special_dir(int directory)

class SpecialDirType:
	DESKTOP = 0
	DOCUMENTS = 1
	DOWNLOAD = 2
	MUSIC = 3
	PICTURES = 4
	PUBLIC_SHARE = 5 
	TEMPLATES = 6
	VIDEOS = 7


def get_special_dir(t):
	cdef char* d
	d = g_get_user_special_dir(t)
	if d == NULL:
		return None
	else:
		return d
	
	
	




# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import Gtk
from blueman.Constants import *
from blueman.Functions import have, dprint, mask_ip4_address
from blueman.Lib import get_net_interfaces, get_net_address, get_net_netmask
from socket import inet_ntoa, inet_aton
from blueman.plugins.ServicePlugin import ServicePlugin

from blueman.main.NetConf import NetConf, DnsMasqHandler, DhcpdHandler
from blueman.main.Config import Config
from blueman.main.Mechanism import Mechanism
from blueman.main.AppletService import AppletService
from random import randint

class Network(ServicePlugin):
	__plugin_info__ = (_("Network"), "gtk-network")
	def on_load(self, container):
		
		self.Builder = Gtk.Builder()
		self.Builder.set_translation_domain("blueman")
		self.Builder.add_from_file(UI_PATH +"/services-network.ui")
		self.widget = self.Builder.get_object("network")
		
		self.ignored_keys = []
		
		container.pack_start(self.widget, True, True, 0)
		
		self.interfaces = []
		for iface in get_net_interfaces():
			if iface != "lo" and iface != "pan1":
				print iface
				ip = inet_aton(get_net_address(iface))
				mask = inet_aton(get_net_netmask(iface))
				self.interfaces.append((iface, ip, mask, mask_ip4_address(ip, mask)))

		self.setup_network()
		try:
			self.ip_check()
		except:
			pass
		return (_("Network"), "gtk-network")
		
	def on_enter(self):
		self.widget.props.visible = True
		
	def on_leave(self):
		self.widget.props.visible = False

	def on_property_changed(self, netconf, key, value):
		dprint(self.ignored_keys)
		if key in self.ignored_keys:
			self.ignored_keys.remove(key)
			return
		if key == "rb_blueman" or key == "dhcp_client":
			if value:
				self.Builder.get_object("rb_blueman").props.active = True
			else:
				self.Builder.get_object("rb_nm").props.active = True
			return
		if key == "rb_nm":
			return
		
		if key == "gn_enable":
			self.Builder.get_object(key).props.active = value
			return
		
		if key == "nap_enable":
			dprint("nap_enable", value)
			self.Builder.get_object(key).props.active = value
			nap_frame = self.Builder.get_object("nap_frame")
			if value:
				nap_frame.props.sensitive = True
			else:
				nap_frame.props.sensitive = False
		
		if key == "ip":
			self.option_changed_notify(key, False)
		else:
			self.option_changed_notify(key)
	
	def on_apply(self):

		if self.on_query_apply_state() == True:
			dprint("network apply")
			

			m = Mechanism()
			nap_enable = self.Builder.get_object("nap_enable")
			if nap_enable.props.active:
				
				r_dnsmasq = self.Builder.get_object("r_dnsmasq")
				if r_dnsmasq.props.active:
					stype = "DnsMasqHandler"
				else:
					stype = "DhcpdHandler"
				
				net_ip = self.Builder.get_object("net_ip")
				net_nat = self.Builder.get_object("net_nat")
				
				try:
					m.EnableNetwork(inet_aton(net_ip.props.text), inet_aton("255.255.255.0"), stype)
					
					if not self.NetConf.props.nap_enable: #race condition workaround
						self.ignored_keys.append("nap_enable")
					self.NetConf.props.nap_enable = True
				except Exception, e:
					lines = str(e).splitlines()
					
					d = Gtk.MessageDialog( None, buttons=Gtk.ButtonsType.OK, type=Gtk.MessageType.ERROR)
					d.props.icon_name = "gtk-dialog-error"
					d.props.text = _("Failed to apply network settings")
					d.props.secondary_text = lines[-1]
					d.run()
					d.destroy()
					return
			else:
				if self.NetConf.props.nap_enable: #race condition workaround
					self.ignored_keys.append("nap_enable")
				self.NetConf.props.nap_enable = False
				m.DisableNetwork()
			
			self.clear_options()

	
	def ip_check(self):
		e = self.Builder.get_object("net_ip")
		address = e.props.text
		try:
			if address.count(".") != 3:
				raise Exception
			a = inet_aton(address)
		except:
			e.props.secondary_icon_stock = Gtk.STOCK_DIALOG_ERROR
			e.props.secondary_icon_tooltip_text = _("Invalid IP address")
			raise
			
		a_netmask = "\xff\xff\xff\0"
		
		a_masked = mask_ip4_address(a, a_netmask)

		for iface, ip, netmask, masked in self.interfaces:
			#print mask_ip4_address(a, netmask).encode("hex_codec"), masked.encode("hex_codec")
			
			if a == ip:
				e.props.secondary_icon_stock = Gtk.STOCK_DIALOG_ERROR
				e.props.secondary_icon_tooltip_text = _("IP address conflicts with interface %s which has the same address" % iface)
				raise Exception
			
			elif mask_ip4_address(a, netmask) == masked:
				e.props.secondary_icon_stock = Gtk.STOCK_DIALOG_WARNING
				e.props.secondary_icon_tooltip_text = _("IP address overlaps with subnet of interface"
					" %s, which has the following configuration %s/%s\nThis may cause incorrect network behavior" % (iface, inet_ntoa(ip), inet_ntoa(netmask)))
				return
		
		e.props.secondary_icon_stock = None
	
	def on_query_apply_state(self):
		changed = False
		opts = self.get_options()
		if opts == []:
			return False
		else:
			if "ip" in opts:
				try:
					self.ip_check()
				except Exception,e:
					print e
					return -1
					
			return True

			
	def setup_network(self):
		self.NetConf = Config("network")
		self.NetConf.connect("property-changed", self.on_property_changed)

		gn_enable = self.Builder.get_object("gn_enable")
		#latest bluez does not support GN, apparently
		gn_enable.props.visible = False
		
		nap_enable = self.Builder.get_object("nap_enable")
		r_dnsmasq = self.Builder.get_object("r_dnsmasq")
		r_dhcpd = self.Builder.get_object("r_dhcpd")
		net_ip = self.Builder.get_object("net_ip")
		net_nat = self.Builder.get_object("net_nat")
		rb_nm = self.Builder.get_object("rb_nm")
		rb_blueman = self.Builder.get_object("rb_blueman")
		rb_dun_nm = self.Builder.get_object("rb_dun_nm")
		rb_dun_blueman = self.Builder.get_object("rb_dun_blueman")
				
		nap_frame = self.Builder.get_object("nap_frame")
		warning = self.Builder.get_object("warning")
		
		rb_blueman.props.active = self.NetConf.props.dhcp_client
		nap_enable.props.active = self.NetConf.props.nap_enable
		gn_enable.props.active = self.NetConf.props.gn_enable
		
		net_ip.props.text = "10.%d.%d.1" % (randint(0, 255), randint(0, 255))
		
		if not self.NetConf.props.nap_enable:
			nap_frame.props.sensitive = False
			
		nc = NetConf.get_default()
		if nc.ip4_address != None:
			net_ip.props.text = inet_ntoa(nc.ip4_address)
			#if not self.NetConf.props.nap_enable:
			#	self.ignored_keys.append("nap_enable")
			self.NetConf.props.nap_enable = True

		
		#if ns["masq"] != 0:
		#	net_nat.props.active = ns["masq"]
			
		if nc.get_dhcp_handler() == None:
			nap_frame.props.sensitive = False
			nap_enable.props.active = False
			if self.NetConf.props.nap_enable or self.NetConf.props.nap_enable == None:
				self.ignored_keys.append("nap_enable")
			self.NetConf.props.nap_enable = False
			
			
		else:
			if nc.get_dhcp_handler() == DnsMasqHandler:
				r_dnsmasq.props.active = True
			else:
				r_dhcpd.props.active = True

		if not have("dnsmasq") and not have("dhcpd3"):
			nap_frame.props.sensitive = False
			warning.props.visible = True
			warning.props.sensitive = True
			nap_enable.props.sensitive = False
			if self.NetConf.props.nap_enable or self.NetConf.props.nap_enable == None:
				self.ignored_keys.append("nap_enable")
			self.NetConf.props.nap_enable = False
			
		if not have("dnsmasq"):
			r_dnsmasq.props.sensitive = False
			r_dnsmasq.props.active = False
			r_dhcpd.props.active = True
		
		if not have("dhcpd3"):
			r_dhcpd.props.sensitive = False
			r_dhcpd.props.active = False
			r_dnsmasq.props.active = True
		
		r_dnsmasq.connect("toggled", lambda x: self.option_changed_notify("dnsmasq"))
		
		net_nat.connect("toggled", lambda x: self.on_property_changed(self.NetConf, "nat", x.props.active))
		net_ip.connect("changed", lambda x: self.on_property_changed(self.NetConf, "ip", x.props.text))
		gn_enable.connect("toggled", lambda x: setattr(self.NetConf.props, "gn_enable", x.props.active))
		nap_enable.connect("toggled", lambda x: self.on_property_changed(self.NetConf, "nap_enable", x.props.active))

		applet = AppletService()
		
		avail_plugins = applet.QueryAvailablePlugins()
		active_plugins = applet.QueryPlugins()
		
		def dun_support_toggled(rb, x):
			if HAL_ENABLED:
				if rb.props.active and x == "nm":
					applet.SetPluginConfig("PPPSupport", False)
					applet.SetPluginConfig("NMIntegration", True)
				elif rb.props.active and x == "blueman":
					applet.SetPluginConfig("NMIntegration", False)
					applet.SetPluginConfig("PPPSupport", True)
			else:
				if rb.props.active and x == "nm":
					applet.SetPluginConfig("PPPSupport", False)
					applet.SetPluginConfig("NMDUNSupport", True)
				elif rb.props.active and x == "blueman":
					applet.SetPluginConfig("NMDUNSupport", False)
					applet.SetPluginConfig("PPPSupport", True)				
				
		def pan_support_toggled(rb, x):
			if rb.props.active and x == "nm":
				applet.SetPluginConfig("DhcpClient", False)
				
				if HAL_ENABLED:
					applet.SetPluginConfig("NMIntegration", True)
				else:
					applet.SetPluginConfig("NMPANSupport", True)
					
			elif rb.props.active and x == "blueman":
				if HAL_ENABLED:
					applet.SetPluginConfig("NMIntegration", False)
				else:
					applet.SetPluginConfig("NMPANSupport", False)
					
				applet.SetPluginConfig("DhcpClient", True)		
	
				
		if "PPPSupport" in active_plugins:
			rb_dun_blueman.props.active = True

		if HAL_ENABLED:
			if not "NMIntegration" in avail_plugins:
				rb_dun_nm.props.sensitive = False
				rb_dun_nm.props.tooltip_text = _("Not currently supported with this setup")

				rb_nm.props.sensitive = False
				rb_nm.props.tooltip_text = _("Not currently supported with this setup")
		else:
			if "NMDUNSupport" in avail_plugins:
				rb_dun_nm.props.sensitive = True
			else:
				rb_dun_nm.props.sensitive = False
				rb_dun_nm.props.tooltip_text = _("Not currently supported with this setup")		
			
			if "NMPANSupport" in avail_plugins:
				rb_nm.props.sensitive = True
			else:
				rb_nm.props.sensitive = False
				rb_nm.props.tooltip_text = _("Not currently supported with this setup")
				
			if "NMPANSupport" in active_plugins:
				rb_nm.props.active = True
		
			if "NMDUNSupport" in active_plugins:
				rb_dun_nm.props.active = True		
				
		rb_nm.connect("toggled", pan_support_toggled, "nm")
		rb_blueman.connect("toggled", pan_support_toggled, "blueman")
				
		rb_dun_nm.connect("toggled", dun_support_toggled, "nm")
		rb_dun_blueman.connect("toggled", dun_support_toggled, "blueman")

				
				
				



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

import dbus
import re
from blueman.Lib import probe_modem
import os
from gi.repository import GObject

def mkname():
    i = 0
    while 1:
        i += 1
        yield "/lib/udev/rules.d/33-%d-blueman.rules" % i
name = mkname()

def getUniqueSynset(hasDupes):
	unique_trick = [ uniq for uniq in hasDupes if uniq not in locals()['_[1]'] ]
	return unique_trick
									        
class HalManager(dbus.proxies.Interface):
	ports = []
	layer = None
	def __init__(self):
		self.bus = dbus.SystemBus()
		obj = self.bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/Manager')
		dbus.proxies.Interface.__init__(self, obj, 'org.freedesktop.Hal.Manager')
		
	
	def register_netdev(self, devicename):
		def reg():
			devices = self.FindDeviceStringMatch("net.interface", devicename)
			if len(devices):
				for device in devices:
					obj = self.bus.get_object('org.freedesktop.Hal', device)
					device = dbus.Interface(obj, 'org.freedesktop.Hal.Device')
					device.SetPropertyString("info.category", "net.80203")
					device.AddCapability("net.80203")
		GObject.timeout_add(1000, reg)

	def register(self, device_file, bd_addr, ok, err):

		existing = self.FindDeviceStringMatch("serial.device", device_file)

		if len(existing) > 0:
			dprint("Already registered")
			ok()
			return	
		
		ref = self.NewDevice()
		r = re.search("rfcomm([0-9]*)$", device_file)
		portid = int(r.groups()[0])
		

		obj = self.bus.get_object('org.freedesktop.Hal', ref)
		device = dbus.Interface(obj, 'org.freedesktop.Hal.Device')

		dev = self.FindDeviceStringMatch("info.product", "Bluetooth RFCOMM")
		if len(dev) == 0:
			dev = self._create_layer()
		else:
			dev = dev[0]
		
		def probe_response(capabilities):
			dprint("Device capabilities: %s" % capabilities)

			if capabilities == None or capabilities == []:
				dprint("Removing temp UDI")
				err(Exception)
			else:
				ok()
				try:
					dprint("Generating udev rule")
					rule_file = name.next()
					rule = open(rule_file, "w")
					rule.write("DEVPATH==\"*/rfcomm%d\", ENV{ID_NM_MODEM_PROBED}=\"1\", " % portid)
					caps = []
					for cap in capabilities:
						if cap == "GSM-07.07" or cap == "GSM-07.05":
							caps.append("ENV{ID_NM_MODEM_GSM}=\"1\"")
						elif cap == "IS-707-A" or cap == "IS-707-P":
							caps.append("ENV{ID_NM_MODEM_IS707_A}=\"1\"")
						elif cap == "IS-856":
							caps.append("ENV{ID_NM_MODEM_IS856}=\"1\"")
						elif cap == "IS-856-A":
							caps.append("ENV{ID_NM_MODEM_IS856_A}=\"1\"")
							
					caps = getUniqueSynset(caps)
					rule.write(", ".join(caps))
					rule.close()
					
				except IOError:
					dprint("Failed to generate udev rule")
				
				else:
					#this causes hal to register a rfcomm device
					uevent = open("/sys/class/tty/rfcomm%s/uevent" % portid, "w")
					uevent.write("change")
					uevent.close()
					
					
				def check_hal(device):
					found_devs = self.FindDeviceStringMatch("linux.device_file", "/dev/rfcomm%d" % portid)
					if len(found_devs):
						self.Remove(found_devs[0])

					dprint("Adding our own rfcomm device to hal")
					device.SetPropertyString("info.category", "serial")
					device.SetPropertyString("info.parent", dev)
					device.SetPropertyString("info.bluetooth_address", bd_addr)
					device.SetPropertyString("info.product", "DUN (%s)" % (bd_addr))
					device.SetPropertyString("info.vendor", "Bluetooth")
					device.SetPropertyString("info.linux.driver", "rfcomm")
					device.SetPropertyString("linux.sysfs_path", "/sys/class/tty/rfcomm%s" % portid) 

					device.SetPropertyInteger("serial.port", portid)
					device.SetPropertyString("serial.device", device_file)
					device.SetPropertyString("linux.device_file", device_file)
					device.SetPropertyString("serial.type", "unknown")

					device.SetPropertyString("serial.originating_device", "/org/freedesktop/Hal/devices/%s" % "rfcomm%s" % portid)

					device.AddCapability("serial")
					device.AddCapability("modem")	
					device.SetMultipleProperties({"modem.command_sets": capabilities})
				
					self.CommitToGdl(ref, "/org/freedesktop/Hal/devices/%s" % "rfcomm%s" % portid)
						
					GObject.timeout_add(1000, os.unlink, rule_file)
		
				GObject.timeout_add(1000, check_hal, device)
		
		dprint("Probing device %s for capabilities" % device_file)
		probe_modem(device_file, probe_response)
		
	def _create_layer(self):
		ref = self.NewDevice()

		obj = self.bus.get_object('org.freedesktop.Hal', ref)
		device = dbus.Interface(obj, 'org.freedesktop.Hal.Device')

		bt_adapter = self.FindDeviceByCapability("bluetooth_hci");

		device.SetPropertyString("info.linux.driver", "rfcomm")
		try:
			device.SetPropertyString("info.parent", bt_adapter[0])
		except:
			pass
			
		device.SetPropertyString("info.product", "Bluetooth RFCOMM")

		self.CommitToGdl(ref, "/org/freedesktop/Hal/devices/bt_rfcomm_layer")
		
		return "/org/freedesktop/Hal/devices/bt_rfcomm_layer"
		
	def remove_all(self):
		devs = self.FindDeviceStringMatch("info.product", "Bluetooth RFCOMM")
		for dev in devs:
			ports = self.FindDeviceStringMatch("info.parent", dev)
			for port in ports:
				self.Remove(port)
			
			self.Remove(dev)
			
			
	def unregister_dev(self, device_file):
		ports = self.FindDeviceStringMatch("serial.device", device_file)
		for port in ports:
			self.Remove(port)
			
	def unregister_addr(self, dev_address):
		ports = self.FindDeviceStringMatch("info.bluetooth_address", device_file)
		for port in ports:
			self.Remove(port)
	
	def exists(self, device_file):
		ports = self.FindDeviceStringMatch("serial.device", device_file)
		return len(ports) > 0
		



import dbus


class BluezDBusException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


class DBusFailedError(BluezDBusException):
    pass


class DBusInvalidArgumentsError(BluezDBusException):
    pass


class DBusNotAuthorizedError(BluezDBusException):
    pass


class DBusOutOfMemoryError(BluezDBusException):
    pass


class DBusNoSuchAdapterError(BluezDBusException):
    pass


class DBusNotReadyError(BluezDBusException):
    pass


class DBusNotAvailableError(BluezDBusException):
    pass


class DBusNotConnectedError(BluezDBusException):
    pass


class DBusConnectionAttemptFailedError(BluezDBusException):
    pass


class DBusAlreadyExistsError(BluezDBusException):
    pass


class DBusDoesNotExistError(BluezDBusException):
    pass


class DBusNoReplyError(BluezDBusException):
    pass


class DBusInProgressError(BluezDBusException):
    pass


class DBusNotSupportedError(BluezDBusException):
    pass


class DBusAuthenticationFailedError(BluezDBusException):
    pass


class DBusAuthenticationTimeoutError(BluezDBusException):
    pass


class DBusAuthenticationRejectedError(BluezDBusException):
    pass


class DBusAuthenticationCanceledError(BluezDBusException):
    pass


class DBusUnsupportedMajorClassError(BluezDBusException):
    pass


class DBusServiceUnknownError(BluezDBusException):
    pass


class DBusMainLoopNotSupportedError(BluezDBusException):
    pass


class DBusMainLoopModuleNotFoundError(BluezDBusException):
    pass


class BluezUnavailableAgentMethodError(BluezDBusException):
    pass


__DICT_ERROR__ = {'org.bluez.Error.Failed:': DBusFailedError,
                  'org.bluez.Error.InvalidArguments:': DBusInvalidArgumentsError,
                  'org.bluez.Error.NotAuthorized:': DBusNotAuthorizedError,
                  'org.bluez.Error.OutOfMemory:': DBusOutOfMemoryError,
                  'org.bluez.Error.NoSuchAdapter:': DBusNoSuchAdapterError,
                  'org.bluez.Error.NotReady:': DBusNotReadyError,
                  'org.bluez.Error.NotAvailable:': DBusNotAvailableError,
                  'org.bluez.Error.NotConnected:': DBusNotConnectedError,
                  'org.bluez.serial.Error.ConnectionAttemptFailed:': DBusConnectionAttemptFailedError,
                  'org.bluez.Error.AlreadyExists:': DBusAlreadyExistsError,
                  'org.bluez.Error.DoesNotExist:': DBusDoesNotExistError,
                  'org.bluez.Error.InProgress:': DBusInProgressError,
                  'org.bluez.Error.NoReply:': DBusNoReplyError,
                  'org.bluez.Error.NotSupported:': DBusNotSupportedError,
                  'org.bluez.Error.AuthenticationFailed:': DBusAuthenticationFailedError,
                  'org.bluez.Error.AuthenticationTimeout:': DBusAuthenticationTimeoutError,
                  'org.bluez.Error.AuthenticationRejected:': DBusAuthenticationRejectedError,
                  'org.bluez.Error.AuthenticationCanceled:': DBusAuthenticationCanceledError,
                  'org.bluez.serial.Error.NotSupported:': DBusNotSupportedError,
                  'org.bluez.Error.UnsupportedMajorClass:': DBusUnsupportedMajorClassError,
                  'org.freedesktop.DBus.Error.ServiceUnknown:': DBusServiceUnknownError}


def parse_dbus_error(exception):
    global __DICT_ERROR__

    aux = "%s" % exception
    aux_splt = aux.split(None, 1)
    try:
        return __DICT_ERROR__[aux_splt[0]](aux_splt[1])
    except KeyError:
        return exception


def raise_dbus_error(func):
    def warp(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except dbus.DBusException as exception:
            raise parse_dbus_error(exception)
    return warp


def raise_type_error(instance, cls):
    if not isinstance(instance, cls):
        raise TypeError('Expecting an instance of ' + cls.__name__ + ', not ' + type(instance).__name__)


from blueman.Functions import dprint
from blueman.plugins.ConfigPlugin import ConfigPlugin
from blueman.main.SignalTracker import SignalTracker
import cPickle as pickle
import os
import atexit
import dbus.service
import dbus.glib
import weakref
import signal
from gi.repository import GObject
import copy


def sighandler():
    print("got signal")
    exit()


signal.signal(signal.SIGTERM, sighandler)
signal.signal(signal.SIGHUP, sighandler)

dbus.service.Object.SUPPORTS_MULTIPLE_OBJECT_PATHS = True

cfg_path = os.path.expanduser('~/.config/blueman/blueman.cfg')


class Monitor(dbus.service.Object):
    __id__ = 0

    def __init__(self, plugin):
        Monitor.__id__ += 1
        self.inst_id = "/inst%d" % Monitor.__id__

        self.plugin = plugin
        self.sigs = SignalTracker()
        self.bus = dbus.SessionBus()

        dbus.service.Object.__init__(self)
        self.add_to_connection(self.bus, self.inst_id)

        self.sigs.Handle("dbus", self.bus, self.on_value_changed, "ValueChanged", "org.blueman.Config")

    def on_value_changed(self, section, data):
        s = "".join(chr(b) for b in data)

        (key, value) = pickle.loads(s)

        key = str(key)

        if self.plugin().section == section:
            self.plugin().set(key, value, True)
        else:
            if not section in File.__db__:
                File.__db__[section] = {}

            File.__db__[section][key] = value

    @dbus.service.signal(dbus_interface="org.blueman.Config", signature='say')
    def ValueChanged(self, section, data):
        pass


class File(ConfigPlugin):
    __priority__ = 1
    __plugin__ = "file"

    __db__ = None

    timeout = None

    def __del__(self):
        self.Monitor.sigs.DisconnectAll()
        self.Monitor.remove_from_connection()
        del self.Monitor

    def on_load(self, section):

        if not File.__db__:
            if not os.path.exists(os.path.expanduser('~/.config/blueman')):
                try:
                    os.makedirs(os.path.expanduser('~/.config/blueman'))
                except:
                    pass
            try:
                f = open(cfg_path, "r")
                File.__db__ = pickle.load(f)
                f.close()
            except Exception as e:
                File.__db__ = {}

            atexit.register(File.save)

        self.config = File.__db__
        self.section = section
        if self.section == "":
            self.section = "__blueman__"

        if not self.section in self.config:
            self.config[self.section] = {}

        self.Monitor = Monitor(weakref.ref(self))

    @staticmethod
    def save():
        dprint("Saving config")
        f = open(cfg_path, "w")
        pickle.dump(File.__db__, f, pickle.HIGHEST_PROTOCOL)
        f.close()
        File.timeout = None

    def set(self, key, value, local=False):

        if key in self.config[self.section]:
            prev = self.config[self.section][key]
        else:
            prev = None

        if prev != value:
            self.config[self.section][key] = copy.deepcopy(value)
            self.emit("property-changed", key, value)

            if not local:
                if File.timeout:
                    GObject.source_remove(File.timeout)
                File.timeout = GObject.timeout_add(1000, File.save)
                self.Monitor.ValueChanged(self.section, pickle.dumps((key, value), pickle.HIGHEST_PROTOCOL))

    def get(self, key):
        if key in self.config[self.section]:
            return copy.deepcopy(self.config[self.section][key])
        else:
            return None

    def list_dirs(self):
        l = []
        for key in self.config.keys():
            if self.section in key:
                k = key.replace(self.section, "")
                s = k.split("/")
                if len(s) > 1:
                    l.append(self.section + "/" + s[1])
        return l



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

class ServicePlugin(object):
	instances = []
	__plugin_info__ = None
	
	def __init__(self, services_inst):
		ServicePlugin.instances.append(self)
		self._options = []
		self._orig_state = {}
		self.__services_inst = services_inst
		
		self.__is_exposed = False
		self._is_loaded = False
		
	def _on_enter(self):
		if not self.__is_exposed:
			self.on_enter()
			self.__is_exposed = True
			
	def _on_leave(self):
		if self.__is_exposed:
			self.on_leave()
			self.__is_exposed = False

	
	#call when option has changed.
	def option_changed_notify(self, option_id, state=True):

		if not option_id in self._options:
			self._options.append(option_id)
		else:
			if state:
				self._options.remove(option_id)

		
		self.__services_inst.option_changed()
	
	def get_options(self):
		return self._options
		
	def clear_options(self):
		self._options = []


	#virtual functions
	#in: container hbox
	#out: (menu entry name, menu icon name)
	def on_load(self, container):
		pass
	
	def on_unload(self):
		pass
	
	#return true if apply button should be sensitive or false if not. -1 to force disabled
	def on_query_apply_state(self):
		pass
		
	def on_apply(self):
		pass
	
	#called when current plugin's page is selected. The plugin's widget should be shown
	def on_enter(self):
		pass
	
	#called when current plugin's page is changed to another. The plugin's widget should be hidden.
	def on_leave(self):
		pass



from gi.repository import Notify
from gi.repository import Gtk
from gi.repository import Gdk
from blueman.Functions import dprint
from blueman.gui.GtkAnimation import AnimBase

Notify.init("blueman")

OPACITY_START = 0.7


class Fade(AnimBase):
    def __init__(self, window):
        AnimBase.__init__(self, state=OPACITY_START)
        self.window = window

    def state_changed(self, state):
        self.window.props.opacity = state


class NotificationDialog(Gtk.MessageDialog):
    def __init__(self, summary, message, timeout=-1, actions=None, actions_cb=None, pixbuf=None, status_icon=None):
        GObject.GObject.__init__(self, parent=None, flags=0, type=Gtk.MessageType.QUESTION,
                                 buttons=Gtk.ButtonsType.NONE, message_format=None)

        self.bubble = NotificationBubble(summary, message, pixbuf=pixbuf)

        i = 100
        self.actions = {}
        self.callback = actions_cb
        if actions:
            for a in actions:
                action_id = a[0]
                action_name = a[1]
                if len(a) == 3:
                    icon_name = a[2]
                else:
                    icon_name = None

                self.actions[i] = action_id
                button = self.add_button(action_name, i)
                if icon_name:
                    im = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
                    im.show()
                    button.props.image = im
                i += 1

        self.actions[Gtk.ResponseType.DELETE_EVENT] = "close"

        self.props.secondary_use_markup = True
        self.resize(350, 50)

        self.fader = Fade(self)

        self.props.skip_taskbar_hint = False

        self.props.title = summary
        self.props.text = summary
        self.props.secondary_text = message

        self.props.window_position = Gtk.WindowPosition.CENTER

        if pixbuf:
            self.set_icon_from_pixbuf(pixbuf)

        self.connect("response", self.dialog_response)
        self.props.icon_name = "blueman"

        self.entered = False

        def on_enter(widget, event):
            if self.window == Gdk.Window.at_pointer()[0] or not self.entered:
                self.fader.animate(start=self.fader.get_state(), end=1.0, duration=500)
                self.entered = True

        def on_leave(widget, event):
            if not Gdk.Window.at_pointer():
                self.entered = False
                self.fader.animate(start=self.fader.get_state(), end=OPACITY_START, duration=500)

        self.connect("enter-notify-event", on_enter)
        self.connect("leave-notify-event", on_leave)

        self.set_opacity(OPACITY_START)
        self.present()
        self.set_opacity(OPACITY_START)


    def get_id(self):
        if self.bubble:
            return self.bubble.props.id

    def dialog_response(self, dialog, response):
        if self.callback:
            self.callback(self, self.actions[response])
        self.hide()

    def close(self):
        self.hide()

    def set_hint_int32(*args):
        dprint("stub")

    def set_timeout(*args):
        dprint("stub")

    def add_action(*args):
        dprint("stub")

    def clear_actions(*args):
        dprint("stub")

    def set_urgency(*args):
        dprint("stub")

    def update(self, summary, message):
        self.props.title = summary

        self.props.text = summary
        self.props.secondary_text = message
        self.present()

    def set_icon_from_pixbuf(self, pixbuf):
        im = Gtk.Image.new_from_pixbuf(pixbuf)
        self.set_image(im)
        im.show()


class NotificationBubble(Notify.Notification):
    def __new__(cls, summary, message, timeout=-1, actions=None, actions_cb=None, pixbuf=None, status_icon=None):
        self = Notify.Notification.new(summary, message, None)

        def on_notification_closed(n, *args):
            self.disconnect(closed_sig)
            if actions_cb:
                actions_cb(n, "closed")

        def on_action(n, action, *args):
            self.disconnect(closed_sig)
            actions_cb(n, action)

        if pixbuf:
            self.set_icon_from_pixbuf(pixbuf)

        if actions:
            for action in actions:
                self.add_action(action[0], action[1], on_action, None)
            self.add_action("default", "Default Action", on_action, None)

        closed_sig = self.connect("closed", on_notification_closed)
        if timeout != -1:
            self.set_timeout(timeout)
        if status_icon:
            _, screen, area, orientation = status_icon.get_geometry()
            self.set_hint_int32("x", area.x + area.width / 2)
            self.set_hint_int32("y", area.y + area.height / 2)

        self.show()

        return self

    def get_id(self):
        return self.props.id


class Notification(object):
    @staticmethod
    def actions_supported():
        return "actions" in Notify.get_server_caps()

    def __new__(cls, summary, message, timeout=-1, actions=None, actions_cb=None, pixbuf=None, status_icon=None):
        if not "actions" in Notify.get_server_caps():
            if actions is not None:
                return NotificationDialog(summary, message, timeout, actions, actions_cb, pixbuf, status_icon)

        return NotificationBubble(summary, message, timeout, actions, actions_cb, pixbuf, status_icon)



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from gi.repository import Pango
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from blueman.Functions import get_icon
from blueman.main.SignalTracker import SignalTracker

class ManagerProgressbar(GObject.GObject):

	__gsignals__ = {
		'cancelled' : (GObject.SignalFlags.RUN_LAST, None, ()),
	}
	__instances__ = []
	def __init__(self, blueman, cancellable=True, text=_("Connecting")):
		def on_enter(evbox, event):
			c = Gdk.Cursor.new(Gdk.CursorType.HAND2)
			self.window.get_window().set_cursor(c)
		
		def on_leave(evbox, event):
			self.window.get_window().set_cursor(None)
			
		def on_clicked(evbox, event):
			self.eventbox.props.sensitive = False
			self.emit("cancelled")
		
		GObject.GObject.__init__(self)
		self.Blueman = blueman
		
		self.cancellable = cancellable
		
		self.hbox = hbox = blueman.Builder.get_object("statusbar1")
		
		self.progressbar = Gtk.ProgressBar()
		
		self.signals = SignalTracker()

		self.button = Gtk.Image.new_from_pixbuf(get_icon("gtk-stop", 16))
	
		self.eventbox = eventbox = Gtk.EventBox()
		eventbox.add(self.button)
		eventbox.props.tooltip_text = _("Cancel Operation")
		self.signals.Handle(eventbox, "enter-notify-event", on_enter)
		self.signals.Handle(eventbox, "leave-notify-event", on_leave)
		self.signals.Handle(eventbox, "button-press-event", on_clicked)

		
		self.progressbar.set_size_request(100, 15)
		self.progressbar.set_ellipsize(Pango.EllipsizeMode.END)
		self.progressbar.set_text(text)
		self.progressbar.set_pulse_step(0.05)
		
		self.window = blueman.Builder.get_object("window")

		hbox.pack_end(eventbox, True, False, 0)
		hbox.pack_end(self.progressbar, False, False, 0)

		if ManagerProgressbar.__instances__ != []:
			dprint("hiding", ManagerProgressbar.__instances__[-1])
			ManagerProgressbar.__instances__[-1].hide()
		
		self.show()
		if not self.cancellable:
			self.eventbox.props.sensitive = False
		
		self.gsource = None
		self.finalized = False
		
		ManagerProgressbar.__instances__.append(self)
		
	def connect(self, *args):
		self.signals.Handle("gobject", super(ManagerProgressbar, self), *args)
		
	def show(self):
		if self.Blueman.Config.props.show_statusbar == False:
			self.Blueman.Builder.get_object("statusbar").props.visible = True

		
		#if self.Blueman.Stats.hbox.size_request()[0] + self.progressbar.size_request()[0] + 16 > self.Blueman.window.get_size()[0]:
		#	self.Blueman.Stats.hbox.hide_all()

		
		self.progressbar.props.visible = True
		self.eventbox.props.visible = True
		self.button.props.visible = True		
	
	def hide(self):
		self.Blueman.Stats.hbox.show_all()		
		self.progressbar.props.visible = False
		self.eventbox.props.visible = False
		self.button.props.visible = False
		
	def message(self, msg, timeout=1500):
		self.stop()
		self.set_label(msg)
		self.set_cancellable(False)
		GObject.timeout_add(timeout, self.finalize)		

		
	def finalize(self):
		if not self.finalized:
			self.hide()
			self.stop()
			Gdk.Window.set_cursor(self.window.get_window(), None)
			self.hbox.remove(self.eventbox)
			self.hbox.remove(self.progressbar)
			#self.hbox.remove(self.seperator)
			self.finalized = True
			
			if ManagerProgressbar.__instances__[-1] == self:
				ManagerProgressbar.__instances__.pop()
				#remove all finalized instances
				for inst in reversed(ManagerProgressbar.__instances__):
					if inst.finalized:
						ManagerProgressbar.__instances__.pop()
					else:
						#show last active progress bar
						inst.show()
						break
						
			if ManagerProgressbar.__instances__ == []:
				if self.Blueman.Config.props.show_statusbar == False:
					self.Blueman.Builder.get_object("statusbar").props.visible = False
					
			self.signals.DisconnectAll()
		
		
	def set_cancellable(self, b, hide=False):
		if b:
			self.eventbox.props.visible = True
			self.eventbox.props.sensitive = True
		else:
			if hide:
				self.eventbox.props.visible = False
			else:
				self.eventbox.props.sensitive = False
		
	def set_label(self, label):
		self.progressbar.props.text = label
		
	def fraction(self, frac):
		if not self.finalized:
			self.progressbar.set_fraction(frac)
			
	def started(self):
		return self.gsource != None
	
	def start(self):
		def pulse():
			self.progressbar.pulse()
			return True
		if not self.gsource:
			self.gsource = GObject.timeout_add(1000/24, pulse)
	
	def stop(self):
		if self.gsource != None:
			GObject.source_remove(self.gsource)
		self.progressbar.set_fraction(0.0)



from gi.repository import GObject
from blueman.Functions import dprint
from blueman.main.SignalTracker import SignalTracker
from blueman.bluez.Adapter import Adapter
from blueman.bluez.Device import Device as BluezDevice
import os
import weakref


class Device(GObject.GObject):
    __gsignals__ = {
        'invalidated': (GObject.SignalFlags.NO_HOOKS, None, ()),
        'property-changed': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, instance):
        GObject.GObject.__init__(self)

        self.Properties = {}
        self.Fake = True
        self.Temp = False

        if isinstance(instance, str) or isinstance(instance, unicode):
            self.Device = BluezDevice(instance)
        else:
            self.Device = instance

        #set fallback icon, fixes lp:#327718
        self.Device.Icon = "blueman"
        self.Device.Class = "unknown"

        self.__services = {}

        self.Valid = True

        self.Signals = SignalTracker()

        dprint("caching initial properties")
        self.Properties = self.Device.get_properties()

        if not "Fake" in self.Properties:
            self.Fake = False

        w = weakref.ref(self)
        if not self.Fake:
            self._obj_path = self.Device.get_object_path()
            self.Signals.Handle("bluez", self.Device, lambda key, value: w() and w().property_changed(key, value),
                                "PropertyChanged")
            object_path = self.Device.get_object_path()
            adapter = Adapter(object_path.replace("/" + os.path.basename(object_path), ""))
            self.Signals.Handle("bluez", adapter, lambda path: w() and w().on_device_removed(path), "DeviceRemoved")

    @property
    def Services(self):
        if len(self.__services) == 0:
            self.init_services()

        return self.__services

    def __del__(self):
        dprint("deleting device", self.get_object_path())
        self.Destroy()

    def get_object_path(self):
        if not self.Fake:
            return self._obj_path

    def on_device_removed(self, path):
        if path == self._obj_path:
            self.emit("invalidated")
            self.Destroy()

    def init_services(self):
        dprint("Loading services")

        if not self.Fake:
            services = self.Device.list_services()
            self.__services = {}
            for service in services:
                name = service.get_interface_name().split(".")
                if name[0] == 'org' and name[1] == 'bluez':
                    name = name[2].lower()
                    if name.endswith('1'):
                        name = name[:-1]
                    self.__services[name] = service

    def Copy(self):
        if not self.Valid:
            raise Exception("Attempted to copy an invalidated device")
        return Device(self.Device)

    def property_changed(self, key, value):
        self.emit("property-changed", key, value)
        self.Properties[key] = value
        if key == "UUIDs":
            self.init_services()

    def Destroy(self):
        dprint("invalidating device", self.get_object_path())
        self.Valid = False
        #self.Device = None
        self.Signals.DisconnectAll()

    #def __del__(self):
    #	dprint("DEBUG: deleting Device instance")

    def get_properties(self):
        #print "Properties requested"
        if not self.Valid:
            raise Exception("Attempted to get properties for an invalidated device")
        return self.Properties

    def __getattr__(self, name):
        if name in self.__dict__["Properties"]:
            if not self.Valid:
                #traceback.print_stack()
                dprint("Warning: Attempted to get %s property for an invalidated device" % name)
            return self.__dict__["Properties"][name]
        else:
            return getattr(self.Device, name)

    def __setattr__(self, key, value):
        if not key in self.__dict__ and "Properties" in self.__dict__ and key in self.__dict__["Properties"]:
            if not self.Valid:
                raise Exception("Attempted to set properties for an invalidated device")
            dprint("Setting property", key, value)
            self.__dict__["Device"].set(key, value)
        else:
            self.__dict__[key] = value



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import GObject
import dbus
from blueman.main.SignalTracker import SignalTracker

class OdsBase(dbus.proxies.Interface, GObject.GObject):
	
#	def OdsMethod(fn):
#		def new(self, *args, **kwargs):

#			fn(self, *args, **kwargs)
#			getattr(super(OdsBase,self), fn.__name__)(*args, **kwargs)
#		return new
	
	def __init__(self, service_name, obj_path):
		self.bus = dbus.SessionBus()
		self._signals = SignalTracker()
		
		service = self.bus.get_object("org.openobex", obj_path)
		GObject.GObject.__init__(self)
		dbus.proxies.Interface.__init__(self, service, service_name)
		
	def DisconnectAll(self):
		self._signals.DisconnectAll()
		
	def Handle(self, signame, handler):
		self._signals.Handle("dbus", self.bus, handler, signame, self.dbus_interface, None, self.object_path)
		
	def GHandle(self, *args):
		self._signals.Handle("gobject", self, *args)




from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
import blueman.main.applet.BluezAgent as BluezAgent
import blueman.bluez as Bluez

from gi.repository import GObject
from gi.repository import Gtk


class AuthAgent(AppletPlugin):
    __description__ = _("Provides passkey, authentication services for BlueZ daemon")
    __icon__ = "gtk-dialog-authentication"
    __author__ = "Walmis"
    __depends__ = ["StatusIcon"]

    def on_load(self, applet):
        self.Applet = applet
        self.add_dbus_method(self.SetTimeHint, in_signature="u")

        self.agents = []
        self.last_event_time = 0

        self.agent_manager = Bluez.AgentManager()

    def SetTimeHint(self, time):
        self.last_event_time = time

    def on_unload(self):
        for agent in self.agents:
            if self.legacy():
                agent.adapter.unregister_agent(agent)
            else:
                self.agent_manager.unregister_agent(agent)

    def on_manager_state_changed(self, state):
        if state:
            adapters = self.Applet.Manager.list_adapters()
            for adapter in adapters:
                self.register_agent(adapter)

        else:
            for agent in self.agents:
                agent.Release()

    def on_adapter_added(self, path):
        adapter = Bluez.Adapter(path)
        self.register_agent(adapter)

    def on_released(self, agent):
        agent.disconnect(agent.signal)
        self.agents.remove(agent)

    def get_event_time(self):
        return self.last_event_time

    def register_agent(self, adapter):
        dprint("Registering agent")
        try:
            if self.legacy():
                agent = BluezAgent.AdapterAgent(self.Applet.Plugins.StatusIcon, adapter, self.get_event_time)
                agent.signal = agent.connect("released", self.on_released)
                adapter.register_agent(agent, "DisplayYesNo")
                self.agents.append(agent)
            elif not self.agents:
                agent = BluezAgent.GlobalAgent(self.Applet.Plugins.StatusIcon, self.get_event_time)
                self.agent_manager.register_agent(agent, "DisplayYesNo", default=True)
                self.agents.append(agent)

        except Exception as e:
            dprint("Failed to register agent")
            dprint(e)

    def legacy(self):
        return self.Applet.Manager.get_interface_version()[0] < 5



from blueman.gui.Notification import Notification
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.Mechanism import Mechanism
from blueman.main.SignalTracker import SignalTracker
from blueman.Functions import *
from blueman.bluez.Network import Network


class DhcpClient(AppletPlugin):
    __description__ = _("Provides a basic dhcp client for Bluetooth PAN connections.")
    __icon__ = "network"
    __author__ = "Walmis"

    def on_load(self, applet):
        self.Signals = SignalTracker()

        self.add_dbus_method(self.DhcpClient, in_signature="s")

        self.Signals.Handle('bluez', Network(), self.on_network_prop_changed, 'PropertyChanged', path_keyword="path")

        self.quering = []

    def on_unload(self):
        self.Signals.DisconnectAll()

    def DhcpClient(self, interface):
        self.dhcp_acquire(interface)

    def on_network_prop_changed(self, key, value, path):
        if key == "Interface":
            if value != "":
                self.dhcp_acquire(value)

    def dhcp_acquire(self, device):
        if device not in self.quering:
            self.quering.append(device)
        else:
            return

        if device != "":
            def reply(ip_address):
                Notification(_("Bluetooth Network"),
                             _("Interface %(0)s bound to IP address %(1)s") % {"0": device, "1": ip_address},
                             pixbuf=get_icon("gtk-network", 48),
                             status_icon=self.Applet.Plugins.StatusIcon)

                self.quering.remove(device)

            def err(msg):
                dprint(msg)
                Notification(_("Bluetooth Network"), _("Failed to obtain an IP address on %s") % (device),
                             pixbuf=get_icon("gtk-network", 48),
                             status_icon=self.Applet.Plugins.StatusIcon)

                self.quering.remove(device)

            Notification(_("Bluetooth Network"), _("Trying to obtain an IP address on %s\nPlease wait..." % device),
                         pixbuf=get_icon("gtk-network", 48),
                         status_icon=self.Applet.Plugins.StatusIcon)

            m = Mechanism()
            m.DhcpClient(device, reply_handler=reply, error_handler=err, timeout=120)



from PropertiesBlueZInterface import PropertiesBlueZInterface
from errors import raise_dbus_error, parse_dbus_error
from Device import Device
import dbus


class Adapter(PropertiesBlueZInterface):
    @raise_dbus_error
    def __init__(self, obj_path=None):
        if self.__class__.get_interface_version()[0] < 5:
            interface = 'org.bluez.Adapter'
        else:
            interface = 'org.bluez.Adapter1'
            proxy = dbus.SystemBus().get_object('org.bluez', '/', follow_name_owner_changes=True)
            self.manager_interface = dbus.Interface(proxy, 'org.freedesktop.DBus.ObjectManager')

        super(Adapter, self).__init__(interface, obj_path)

    @raise_dbus_error
    def find_device(self, address):
        devices = self.list_devices()
        for device in devices:
            if device.get_properties()['Address'] == address:
                return device

    @raise_dbus_error
    def list_devices(self):
        if self.__class__.get_interface_version()[0] < 5:
            devices = self.get_interface().ListDevices()
            return [Device(device) for device in devices]
        else:
            objects = self.manager_interface.GetManagedObjects()
            devices = []
            for path, interfaces in objects.items():
                if 'org.bluez.Device1' in interfaces:
                    devices.append(path)
            return [Device(device) for device in devices]

    @raise_dbus_error
    def create_paired_device(self, address, agent_path, capability, error_handler=None, reply_handler=None, timeout=120):
        # BlueZ 4 only!
        def reply_handler_wrapper(obj_path):
            if not callable(reply_handler):
                return
            reply_handler(Device(obj_path))

        def error_handler_wrapper(exception):
            exception = parse_dbus_error(exception)
            if not callable(error_handler):
                raise exception
            error_handler(exception)

        if reply_handler is None and error_handler is None:
            obj_path = self.get_interface().CreatePairedDevice(address, agent_path, capability)
            return Device(obj_path)
        else:
            self.get_interface().CreatePairedDevice(address, agent_path, capability,
                                                    reply_handler=reply_handler_wrapper,
                                                    error_handler=error_handler_wrapper, timeout=timeout)

    @raise_dbus_error
    def create_device(self, address, reply_handler=None, error_handler=None):
        # BlueZ 4 only!
        def reply_handler_wrapper(path):
            if not callable(reply_handler):
                return
            reply_handler(Device(path))

        def error_handler_wrapper(exception):
            exception = parse_dbus_error(exception)
            if not callable(error_handler):
                raise exception
            error_handler(exception)

        if reply_handler is None and error_handler is None:
            obj_path = self.get_interface().CreateDevice(address)
            return Device(obj_path)
        else:
            self.get_interface().CreateDevice(
                address, reply_handler=reply_handler_wrapper, error_handler=error_handler_wrapper
            )
            return None

    @raise_dbus_error
    def handle_signal(self, handler, signal, **kwargs):
        if signal in ['DeviceCreated', 'DeviceRemoved', 'DeviceFound']:
            if self.__class__.get_interface_version()[0] < 5:
                self._handle_signal(handler, signal, self.get_interface_name(), self.get_object_path(), **kwargs)
            else:
                if signal == 'DeviceFound':
                    return

                def wrapper(object_path, interfaces):
                    if 'org.bluez.Device1' in interfaces:
                        handler(object_path)

                self._handler_wrappers[handler] = wrapper

                signal = {
                    'DeviceCreated': 'InterfacesAdded',
                    'DeviceRemoved': 'InterfacesRemoved'
                }[signal]

                self._handle_signal(wrapper, signal, 'org.freedesktop.DBus.ObjectManager', '/', **kwargs)
        else:
            super(Adapter, self).handle_signal(handler, signal, **kwargs)

    @raise_dbus_error
    def unhandle_signal(self, handler, signal, **kwargs):
        if signal in ['DeviceCreated', 'DeviceRemoved', 'DeviceFound']:
            if self.__class__.get_interface_version()[0] < 5:
                self._unhandle_signal(handler, signal, self.get_interface_name(), self.get_object_path(), **kwargs)
            else:
                if signal == 'DeviceFound':
                    return

                signal = {
                    'DeviceCreated': 'InterfacesAdded',
                    'DeviceRemoved': 'InterfacesRemoved'
                }[signal]

                self._unhandle_signal(self._handler_wrappers[handler], signal, self.get_interface_name(),
                                      self.get_object_path(), **kwargs)
        else:
            super(Adapter, self).unhandle_signal(handler, signal, **kwargs)

    @raise_dbus_error
    def start_discovery(self):
        self.get_interface().StartDiscovery()

    @raise_dbus_error
    def stop_discovery(self):
        self.get_interface().StopDiscovery()

    @raise_dbus_error
    def remove_device(self, device):
        self.get_interface().RemoveDevice(device.get_object_path())

    @raise_dbus_error
    def register_agent(self, agent, capability=''):
        # BlueZ 4 only!
        self.get_interface().RegisterAgent(agent.get_object_path(), capability)

    @raise_dbus_error
    def unregister_agent(self, agent):
        # BlueZ 4 only!
        self.get_interface().UnregisterAgent(agent.get_object_path())



from blueman.plugins.ManagerPlugin import ManagerPlugin
from blueman.main.PulseAudioUtils import PulseAudioUtils, EventType
from blueman.main.SignalTracker import SignalTracker
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu
from blueman.gui.MessageArea import MessageArea
from blueman.Functions import get_icon, create_menuitem

from gi.repository import Gtk

class PulseAudioProfile(ManagerPlugin):
    def on_load(self, user_data):
        self.devices = {}
        self.item = None

        self.deferred = []

        pa = PulseAudioUtils()
        pa.connect("event", self.on_pa_event)
        pa.connect("connected", self.on_pa_ready)

    def on_pa_ready(self, utils):
        dprint("connected")
        for dev in self.deferred:
            self.regenerate_with_device(dev.Address)

        self.deferred = []


    #updates all menu instances with the following device address
    def regenerate_with_device(self, device_addr):
        for inst in ManagerDeviceMenu.__instances__:
            if inst.SelectedDevice.Address == device_addr and not inst.is_popup:
                inst.Generate()

    def on_pa_event(self, utils, event, idx):
        dprint(event, idx)

        def get_card_cb(card):
            if card["driver"] == "module-bluetooth-device.c":
                self.devices[card["proplist"]["device.string"]] = card
                self.regenerate_with_device(card["proplist"]["device.string"])

        if event & EventType.CARD:
            print "card",
            if event & EventType.CHANGE:
                print "change"
                utils.GetCard(idx, get_card_cb)
            elif event & EventType.REMOVE:
                print "remove"
            else:
                print "add"
                utils.GetCard(idx, get_card_cb)

    def is_connected(self, device):
        props = device.get_properties()
        if 'Connected' in props:
            return props['Connected']

        # BlueZ 4 only

        try:
            s = device.Services["audiosink"]
            props = s.get_properties()
            if props["Connected"]:
                return True
        except KeyError:
            pass

        try:
            s = device.Services["audiosource"]
            props = s.get_properties()
            if props["State"] != "disconnected":
                return True
        except KeyError:
            pass

        try:
            s = device.Services["headset"]
            props = s.get_properties()
            if props["State"] != "disconnected":
                return True
        except KeyError:
            pass

        return False

    def query_pa(self, device):
        def list_cb(cards):
            for c in cards.itervalues():
                if c["proplist"]["device.string"] == device.Address:
                    self.devices[device.Address] = c
                    self.generate_menu(device)
                    return

        pa = PulseAudioUtils()
        pa.ListCards(list_cb)

    def on_selection_changed(self, item, device, profile):
        if item.get_active():
            pa = PulseAudioUtils()

            c = self.devices[device.Address]

            def on_result(res):
                if not res:
                    MessageArea.show_message(_("Failed to change profile to %s" % profile))

            pa.SetCardProfile(c["index"], profile, on_result)

    def generate_menu(self, device):
        info = self.devices[device.Address]
        group = []

        sub = Gtk.Menu()

        if info:
            for profile in info["profiles"]:
                i = Gtk.RadioMenuItem.new_with_label(group, profile["description"])
                group = i.get_group()

                if profile["name"] == info["active_profile"]:
                    i.set_active(True)

                i.connect("toggled", self.on_selection_changed,
                          device, profile["name"])

                sub.append(i)
                i.show()

            self.item.set_submenu(sub)
            self.item.show()

    def on_request_menu_items(self, manager_menu, device):

        if self.is_connected(device):
            pa = PulseAudioUtils()
            if not pa.connected:
                self.deferred.append(device)
                return

            self.item = create_menuitem(_("Audio Profile"), get_icon("audio-card", 16))
            self.item.props.tooltip_text = _("Select audio profile for PulseAudio")

            if not device.Address in self.devices:
                self.query_pa(device)
            else:
                self.generate_menu(device)

        else:
            return

        return [(self.item, 300)]



from gi.repository import Gtk
from gi.repository import GObject


# noinspection PyAttributeOutsideInit
class GenericList(Gtk.TreeView):
    def __init__(self, data):
        GObject.GObject.__init__(self)
        self.selection = self.get_selection()
        self._load(data)

    def _load(self, data):
        self.ids = {}
        self.columns = {}

        types = []
        for i in range(len(data)):
            types.append(data[i][1])

        types = tuple(types)

        self.liststore = Gtk.ListStore(*types)
        self.set_model(self.liststore)

        for i in range(len(data)):

            self.ids[data[i][0]] = i

            if len(data[i]) == 5 or len(data[i]) == 6:

                column = Gtk.TreeViewColumn(data[i][4])

                column.pack_start(data[i][2], True)
                column.set_attributes(data[i][2], **data[i][3])

                if len(data[i]) == 6:
                    column.set_properties(**data[i][5])

                self.columns[data[i][0]] = column
                self.append_column(column)

    def selected(self):
        (model, iter) = self.selection.get_selected()

        return iter

    def delete(self, id):
        if type(id) == Gtk.TreeIter:
            iter = id
        else:
            iter = self.get_iter(id)

        if iter is None:
            return False
        if self.liststore.iter_is_valid(iter):
            self.liststore.remove(iter)
            return True
        else:
            return False

    def _add(self, **columns):
        items = {}
        for k, v in self.ids.items():
            items[v] = None

        for k, v in columns.items():
            if k in self.ids:
                items[self.ids[k]] = v
            else:
                raise Exception("Invalid key %s" % k)

        return items.values()

    def append(self, **columns):
        vals = self._add(**columns)
        return self.liststore.append(vals)

    def prepend(self, **columns):
        vals = self._add(**columns)
        return self.liststore.prepend(vals)

    def get_conditional(self, **cols):
        ret = []
        matches = 0
        for i in range(len(self.liststore)):
            row = self.get(i)
            for k, v in cols.items():
                if row[k] == v:
                    matches += 1

            if matches == len(cols):
                ret.append(i)

            matches = 0

        return ret

    def set(self, id, **cols):
        if type(id) == Gtk.TreeIter:
            iter = id
        else:
            iter = self.get_iter(id)

        if iter is not None:
            for k, v in cols.items():
                self.liststore.set(iter, self.ids[k], v)
            return True
        else:
            return False

    def get(self, id, *items):
        ret = {}

        if id is not None:
            if type(id) == Gtk.TreeIter:
                iter = id
            else:
                iter = self.get_iter(id)
            if len(items) == 0:
                for k, v in self.ids.items():
                    ret[k] = self.liststore.get(iter, v)[0]
            else:
                for i in range(len(items)):
                    if items[i] in self.ids:
                        ret[items[i]] = self.liststore.get(iter, self.ids[items[i]])[0]
        else:
            return False

        return ret

    def get_iter(self, path):
        if path is None:
            return None

        try:
            return self.liststore.get_iter(path)
        except:
            return None

    def clear(self):
        self.liststore.clear()

    def compare(self, iter_a, iter_b):
        if iter_a is not None and iter_b is not None:
            return self.get_model().get_path(iter_a) == self.get_model().get_path(iter_b)
        else:
            return False



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import Gtk
from gi.repository import GObject


class CellRendererPixbufTable(Gtk.CellRenderer):
	__gproperties__ = {
		"pixbuffs": (GObject.TYPE_PYOBJECT, "pixbuf", "pixbuf", GObject.PARAM_READWRITE)

	}

	def __init__(self):
		self.__gobject_init__()


		self.set_property("yalign", 0.5)
		self.set_property("xalign", 0.5)

	
	def do_set_property(self, pspec, value):
			setattr(self, pspec.name, value)

	def do_get_property(self, pspec):
			return getattr(self, pspec.name)

	def on_render(self, window, widget, background_area, cell_area, expose_area, flags):
		if not self.pixbuffs or self.pixbuffs.cols == 0:
			return


		pix_rect = ()
		pix_rect.x, pix_rect.y, pix_rect.width, pix_rect.height = self.on_get_size(widget, cell_area)

		
		pix_rect.x += cell_area.x
		pix_rect.y += cell_area.y
		pix_rect.width  -= 2 * self.get_property("xpad") + (self.pixbuffs.total_width - self.pixbuffs.size)
		pix_rect.height -= 2 * self.get_property("ypad") + (self.pixbuffs.total_height - self.pixbuffs.size)

		
		row = 0
		col = 0
		
	
		for k,v in self.pixbuffs.get().iteritems():
			#print rows
			if row == self.pixbuffs.rows:
				y_space = 0
				row=0
				col+=1

			else:
				y_space = self.pixbuffs.spacingy
				
			if col == 0 or col == self.pixbuffs.cols:
				x_space = 0
			else:
				x_space = self.pixbuffs.spacingx
		
				

			draw_rect = cell_area.intersect(pix_rect)
			draw_rect = expose_area.intersect(draw_rect)
			
			
			if self.pixbuffs.cols > 2:
				z = self.pixbuffs.size*(self.pixbuffs.cols-1)
			else:
				z = 0
			
			h = v.get_height()
			w = v.get_width()
			#if w > h:
			#	x = 
				
			
			
			window.draw_pixbuf(
					widget.style.black_gc, 
					v, 
					draw_rect.x - pix_rect.x, #source x
					draw_rect.y - pix_rect.y, #source y
					int(draw_rect.x + self.pixbuffs.size * col + x_space*col + (cell_area.width-self.pixbuffs.total_width) * self.get_property("xalign") + (h - w)/2), #dest x
					int(draw_rect.y + self.pixbuffs.size * row + y_space*row + (cell_area.height-self.pixbuffs.total_height) * self.get_property("yalign")), #dest y
					-1,
					-1,
					Gdk.RGB_DITHER_NONE,
					0,
					0
					)
			
			row += 1


	def on_get_size(self, widget, cell_area):
		if not self.pixbuffs or self.pixbuffs.cols == 0:
			return 0, 0, 0, 0


		calc_width  = self.get_property("xpad") * 2 + self.pixbuffs.size + (self.pixbuffs.total_width - self.pixbuffs.size)
		calc_height = self.get_property("ypad") * 2 + self.pixbuffs.size + (self.pixbuffs.total_height - self.pixbuffs.size)
		x_offset = 0
		y_offset = 0
		
		return x_offset, y_offset, calc_width, calc_height

GObject.type_register(CellRendererPixbufTable)



# Copyright (C) 2011 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2011 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

#from blueman.Functions import *

from blueman.Constants import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.applet.BluezAgent import AdapterAgent
import blueman.bluez as Bluez

from gi.repository import GObject
import os
from gi.repository import AppIndicator3 as girAppIndicator

class AppIndicator(AppletPlugin):
    __description__ = _("Uses libappindicator to show a statusicon")
    __icon__ = "blueman"
    __author__ = "Walmis"
    __depends__ = ["StatusIcon", "Menu"]
    
    def on_load(self, applet):
        
        self.indicator = girAppIndicator.Indicator.new ("blueman",
                                                        applet.Plugins.StatusIcon.get_option("icon"),
                                                        girAppIndicator.IndicatorCategory.APPLICATION_STATUS)
        
        self.indicator.set_status (girAppIndicator.IndicatorStatus.ACTIVE)
        
        self.indicator.set_menu(applet.Plugins.Menu.get_menu())
        
        self.s =  self.Applet.Plugins.StatusIcon.connect("notify::icon-name", self.on_notify)

        #self.Applet.Plugins.StatusIcon.set_visible = partial(self.set_visible, self.Applet.Plugins.StatusIcon)
        self.override_method(self.Applet.Plugins.StatusIcon, "set_visible", self.set_visible)
        
        self.Applet.Plugins.StatusIcon.props.visible = False

    def set_visible(self, statusicon, visible):
        if visible:
            self.indicator.set_status(girAppIndicator.IndicatorStatus.ACTIVE)
        else:
            self.indicator.set_status(girAppIndicator.IndicatorStatus.PASSIVE)
        
    def on_notify(self, *args):
        self.update_icon()
        
    def on_unload(self):
        #self.indicator.set_menu(None)
        del self.indicator
        self.Applet.Plugins.StatusIcon.QueryVisibility()
        self.Applet.Plugins.StatusIcon.disconnect(self.s)
        
    #def on_status_icon_pixbuf_ready(self):
    #    print "aaaaa"
    #    GObject.idle_add(self.update_icon)
        
    def update_icon(self):
        
        self.indicator.set_icon(self.Applet.Plugins.StatusIcon.props.icon_name)
        self.indicator.set_status(girAppIndicator.IndicatorStatus.ATTENTION)
        self.indicator.set_status(girAppIndicator.IndicatorStatus.ACTIVE)
        



#
#
# blueman
# (c) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from blueman.gui.DeviceList import DeviceList
from blueman.gui.PixbufTable import PixbufTable
from blueman.gui.CellRendererPixbufTable import CellRendererPixbufTable
from blueman.DeviceClass import get_minor_class, get_major_class
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu
from blueman.Sdp import *

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango
from blueman.Constants import *
from blueman.Functions import *
import cgi

from blueman.gui.GtkAnimation import TreeRowColorFade, TreeRowFade, CellFade

class ManagerDeviceList(DeviceList):
	
	def __init__(self, adapter=None, inst=None):
		cr = Gtk.CellRendererText()
		cr.props.ellipsize = Pango.EllipsizeMode.END
		data = [
			#device picture
			["device_pb", 'GdkPixbuf', Gtk.CellRendererPixbuf(), {"pixbuf":0}, None],
			
			#device caption
			["caption", str, cr, {"markup":1}, None, {"expand": True}],

			
			["rssi_pb", 'GdkPixbuf', Gtk.CellRendererPixbuf(), {"pixbuf":2}, None, {"spacing": 0}],
			["lq_pb", 'GdkPixbuf', Gtk.CellRendererPixbuf(), {"pixbuf":3}, None, {"spacing": 0}],
			["tpl_pb", 'GdkPixbuf', Gtk.CellRendererPixbuf(), {"pixbuf":4}, None, {"spacing": 0}],
			
			#trusted/bonded icons
			#["tb_icons", 'PyObject', CellRendererPixbufTable(), {"pixbuffs":5}, None],
			
			["connected", bool], #used for quick access instead of device.GetProperties
			["bonded", bool], #used for quick access instead of device.GetProperties
			["trusted", bool], #used for quick access instead of device.GetProperties	
			["fake", bool], #used for quick access instead of device.GetProperties, 
					#fake determines whether device is "discovered" or a real bluez device
			
			["rssi", float],
			["lq", float],
			["tpl", float],
			["orig_icon", 'GdkPixbuf'],
			["cell_fader", GObject.TYPE_PYOBJECT],
			["row_fader", GObject.TYPE_PYOBJECT],
			["levels_visible", bool],
			["initial_anim", bool],
		]
		DeviceList.__init__(self, adapter, data)
		self.set_headers_visible(False)
		self.props.has_tooltip = True
		self.Blueman = inst
		
		self.connect("query-tooltip", self.tooltip_query)		
		self.tooltip_row = None
		self.tooltip_col = None
		
		self.connect("button_press_event", self.on_event_clicked)
		self.connect("button_release_event", self.on_event_clicked)
		
		self.menu = None
		
		self.connect("drag_data_received", self.drag_recv)
		self.connect("drag-motion", self.drag_motion)

		Gtk.Widget.drag_dest_set(self, Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY|Gdk.DragAction.DEFAULT)
		Gtk.Widget.drag_dest_add_uri_targets(self)
		
		self.set_search_equal_func(self.search_func, None)
		
	def do_device_found(self, device):
		iter = self.find_device(device)
		if iter:
			anim = TreeRowColorFade(self, self.props.model.get_path(iter), Gdk.color_parse("blue"))
			anim.animate(start=0.8, end=1.0)
		
		
	def search_func(self, model, column, key, iter):
		row = self.get(iter, "caption")
		if key.lower() in row["caption"].lower():
			return False
		print model, column, key, iter
		return True	
	
	def drag_recv(self, widget, context, x, y, selection, target_type, time):

		uris = selection.get_uris()
		uris = list(uris)
		
		context.finish(True, False, time)
		
		path = self.get_path_at_pos(x, y)
		if path:
			iter = self.get_iter(path[0])
			device = self.get(iter, "device")["device"]
			uris.insert(0, "blueman-sendto")
			uris.insert(1, "--device=%s" % device.Address)
			
			spawn(uris)
			context.finish(True, False, time)
		else:
			context.finish(False, False, time)
		
		return True

	
	def drag_motion(self, widget, drag_context, x, y, timestamp):
		path = self.get_path_at_pos(x, y)
		if path != None:
			if path[0] != self.selected():
				iter = self.get_iter(path[0])
				device = self.get(iter, "device")["device"]
				if not device.Fake:	
					found = False
					for uuid in device.UUIDs:
						uuid16 = uuid128_to_uuid16(uuid)
						if uuid16 == OBEX_OBJPUSH_SVCLASS_ID:
							found = True
							break
					if found:
						drag_context.drag_status(Gdk.DragAction.COPY, timestamp)
						self.set_cursor(path[0])
						return True
					else:
						drag_context.drag_status(Gdk.DragAction.DEFAULT, timestamp)
						return False
				else:
					drag_context.drag_status(Gdk.DragAction.COPY, timestamp)
					self.set_cursor(path[0])
					return True
		else:
			drag_context.drag_status(Gdk.DragAction.DEFAULT, timestamp)
			return False			

	
	def on_event_clicked(self, widget, event):

		if event.type==Gdk.EventType.BUTTON_PRESS and event.button==3:
			path = self.get_path_at_pos(int(event.x), int(event.y))
			if path != None:
				row = self.get(path[0], "device")
			
				if row:
					device = row["device"]
					if self.Blueman != None:
						if self.menu == None:
							self.menu = ManagerDeviceMenu(self.Blueman)
					
						self.menu.popup(None, None, None, None, event.button, event.time)

	
	
	
	def get_device_icon(self, klass):
		return get_icon("blueman-"+klass.replace(" ", "-").lower(), 48, "blueman")
		
	


	def make_device_icon(self, target, is_bonded=False, is_trusted=False, is_discovered=False, opacity=255):
		if opacity != 255:
			target = opacify_pixbuf(target, opacity)
		
		sources = []
		if is_bonded:
			sources.append((get_icon("gtk-dialog-authentication", 16), 0, 0, 200))
		
		if is_trusted:
			sources.append((get_icon("blueman-trust", 16), 0, 32, 200))
	
		if is_discovered:
			sources.append((get_icon("gtk-find", 24), 24, 0, 255))

		return composite_icon(target, sources)
	
	
	def device_remove_event(self, device, iter):
		if device.Temp:
			DeviceList.device_remove_event(self, device, iter)
		else:
			row_fader = self.get(iter, "row_fader")["row_fader"]
		
			def on_finished(fader):
			
				fader.disconnect(signal)
				fader.freeze()
				DeviceList.device_remove_event(self, device, iter)	
		
			signal = row_fader.connect("animation-finished", on_finished)
			row_fader.thaw()
			self.emit("device-selected", None, None)
			row_fader.animate(start=row_fader.get_state(), end=0.0, duration=400)
	
	def device_add_event(self, device):
		if device.Fake:
			self.PrependDevice(device)
			GObject.idle_add(self.props.vadjustment.set_value , 0)
			return

		if self.Blueman.Config.props.latest_last:
			self.AppendDevice(device)
		else:
			self.PrependDevice(device)
		
	def make_caption(self, name, klass, address):
		return "<span size='x-large'>%(0)s</span>\n<span size='small'>%(1)s</span>\n<i>%(2)s</i>" % {"0":cgi.escape(name), "1":klass.capitalize(), "2":address}
		
	def get_device_class(self, device):
		klass = get_minor_class(device.Class)
		if klass != "uncategorized":
			return get_minor_class(device.Class, True)
		else:
			return get_major_class(device.Class)		
	
	def row_setup_event(self, iter, device):
		if not self.get(iter, "initial_anim")["initial_anim"]:
			
			cell_fader = CellFade(self, self.props.model.get_path(iter), [2,3,4])
			row_fader = TreeRowFade(self, self.props.model.get_path(iter))
		
		
			self.set(iter, row_fader=row_fader, cell_fader=cell_fader, levels_visible=False)
		
			cell_fader.freeze()		
			def on_finished(fader):
				fader.disconnect(signal)
				fader.freeze()
		
			signal = row_fader.connect("animation-finished", on_finished)
			row_fader.set_state(0.0)
			row_fader.animate(start=0.0, end=1.0, duration=500)
			
			self.set(iter, initial_anim=True)
		

		klass = get_minor_class(device.Class)
		if klass != "uncategorized":
			icon = self.get_device_icon(klass)
			#get translated version
			klass = get_minor_class(device.Class, True)
		else:
			icon = get_icon(device.Icon, 48, "blueman")
			klass = get_major_class(device.Class)

		name = device.Alias
		address = device.Address
		
		caption = self.make_caption(name, klass, address)
		
		#caption = "<span size='x-large'>%(0)s</span>\n<span size='small'>%(1)s</span>\n<i>%(2)s</i>" % {"0":name, "1":klass.capitalize(), "2":address}
		self.set(iter, caption=caption, orig_icon=icon)
		
		self.row_update_event(iter, "Fake", device.Fake)
		
		try:
			self.row_update_event(iter, "Trusted", device.Trusted)
		except:
			pass
		try:
			self.row_update_event(iter, "Paired", device.Paired)
		except:
			pass
			
				
	def row_update_event(self, iter, key, value):
		dprint("row update event", key, value)
		
		#this property is only emitted when device is fake
		if key == "RSSI":
			row = self.get(iter, "orig_icon")
			
			#minimum opacity 90
			#maximum opacity 255
			#rssi at minimum opacity -100
			#rssi at maximum opacity -45
			# y = kx + b
			#solve linear system
			#{ 90 = k * -100 + b
			#{ 255 = k * -45 + b
			# k = 3
			# b = 390
			#and we have a formula for opacity based on rssi :)
			opacity = int(3 * value + 390)
			if opacity > 255:
				opacity = 255
			if opacity < 90:
				opacity = 90
			print "opacity", opacity
			icon = self.make_device_icon(row["orig_icon"], is_discovered=True, opacity=opacity) 
			self.set(iter, device_pb=icon)
		
		elif key == "Trusted":
			row = self.get(iter, "bonded", "orig_icon")
			if value:
				icon = self.make_device_icon(row["orig_icon"], row["bonded"], True, False) 
				self.set(iter, device_pb=icon, trusted=True)
			else:
				icon = self.make_device_icon(row["orig_icon"], row["bonded"], False, False) 
				self.set(iter, device_pb=icon, trusted=False)
			
		
		elif key == "Paired":
			row = self.get(iter, "trusted", "orig_icon")
			if value:
				icon = self.make_device_icon(row["orig_icon"], True, row["trusted"], False) 
				self.set(iter, device_pb=icon, bonded=True)
			else:
				icon = self.make_device_icon(row["orig_icon"], False, row["trusted"], False) 
				self.set(iter, device_pb=icon, bonded=False)
				
		elif key == "Fake":
			row = self.get(iter, "bonded", "trusted", "orig_icon")
			if value:
				icon = self.make_device_icon(row["orig_icon"], False, False, True) 
				self.set(iter, device_pb=icon, fake=True)
			else:
				icon = self.make_device_icon(row["orig_icon"], row["bonded"], row["trusted"], False) 
				self.set(iter, device_pb=icon, fake=False)
				
		elif key == "Alias" or key == "Class":
			device = self.get(iter, "device")["device"]
			c = self.make_caption(value, self.get_device_class(device), device.Address)
			self.set(iter, caption=c)
				
	
	def level_setup_event(self, row_ref, device, cinfo):
		def rnd(value):
			return int(round(value,-1))
		if not row_ref.valid():	
			return
			
		iter = self.get_iter(row_ref.get_path())
		if True:
			if cinfo != None:
				try:
					rssi = float(cinfo.get_rssi())
				except:
					rssi = 0
				try:
					lq = float(cinfo.get_lq())
				except:
					lq = 0
				
				try:
					tpl = float(cinfo.get_tpl())
				except:
					tpl = 0

				rssi_perc = 50 + (rssi / 127 / 2 * 100)
				tpl_perc = 50 + (tpl / 127 / 2 * 100)
				lq_perc = lq / 255 * 100
				
				if lq_perc < 10:
					lq_perc = 10
				if rssi_perc < 10:
					rssi_perc = 10
				if tpl_perc < 10:
					tpl_perc = 10

				
				row = self.get(iter, "levels_visible", "cell_fader", "rssi", "lq", "tpl")
				if not row["levels_visible"]:
					dprint("animating up")
					self.set(iter, levels_visible=True)
					fader = row["cell_fader"]
					fader.thaw()
					fader.set_state(0.0)
					fader.animate(start=0.0, end=1.0, duration=400)
					
					def on_finished(fader):
						fader.freeze()
						fader.disconnect(signal)
						
					signal = fader.connect("animation-finished", on_finished )
					
				if rnd(row["rssi"]) != rnd(rssi_perc):
					self.set(iter, rssi_pb=get_icon("blueman-rssi-%s" % rnd(rssi_perc), 48))
				
				if rnd(row["lq"]) != rnd(lq_perc):
					self.set(iter, lq_pb=get_icon("blueman-lq-%s" % rnd(lq_perc), 48))
				
				if rnd(row["tpl"]) != rnd(tpl_perc):
					self.set(iter, tpl_pb=get_icon("blueman-tpl-%s" % rnd(tpl_perc), 48))
					
					
				self.set(iter,
						rssi=rssi_perc,
						lq=lq_perc,
						tpl=tpl_perc,
						connected=True)
			else:
				
				row = self.get(iter, "levels_visible", "cell_fader")
				if row["levels_visible"]:
					dprint("animating down")
					self.set(iter, levels_visible=False,
						rssi=-1,
						lq=-1,
						tpl=-1)
					fader = row["cell_fader"]
					fader.thaw()
					fader.set_state(1.0)
					fader.animate(start=fader.get_state(), end=0.0, duration=400)
					def on_finished(fader):
						fader.disconnect(signal)
						fader.freeze()
						if row_ref.valid():
							self.set(iter, rssi_pb=None, lq_pb=None, tpl_pb=None, connected=False)
					
					signal = fader.connect("animation-finished", on_finished)

				
		else:
			dprint("invisible")
		
	
	def tooltip_query(self, tw, x, y, kb, tooltip):

		#print args
		#args[4].set_text("test"+str(args[1]))

		path = self.get_path_at_pos(x, y)

		if path is not None:
			if path[0] != self.tooltip_row or path[1] != self.tooltip_col:
				self.tooltip_row = path[0]
				self.tooltip_col = path[1]
				return False
			
			if path[1] == self.columns["device_pb"]:
				iter = self.get_iter(path[0])
				
				row = self.get(iter, "trusted", "bonded")
				trusted = row["trusted"]
				bonded = row["bonded"]
				if trusted and bonded:
					tooltip.set_markup(_("<b>Trusted and Bonded</b>"))
				elif bonded:
					tooltip.set_markup(_("<b>Bonded</b>"))
				elif trusted:
					tooltip.set_markup(_("<b>Trusted</b>"))
				else:
					return False
				
					
				self.tooltip_row = path[0]
				self.tooltip_col = path[1]
				return True

			
			if path[1] == self.columns["tpl_pb"] or path[1] == self.columns["lq_pb"] or path[1] == self.columns["rssi_pb"]:
				iter = self.get_iter(path[0])

				dt = self.get(iter, "connected")["connected"]
				#print dt
				if dt:
					rssi = self.get(iter, "rssi")["rssi"]
					lq = self.get(iter, "lq")["lq"]
					tpl = self.get(iter, "tpl")["tpl"]
					
					if rssi < 30:
						rssi_state = _("Poor")
					
					if rssi < 40 and rssi > 30:
						rssi_state = _("Sub-optimal")

					elif rssi > 40 and rssi < 60:
						rssi_state = _("Optimal")
					
					elif rssi > 60:
						rssi_state = _("Much")
					
					elif rssi > 70:
						rssi_state = _("Too much")
						
					
					if tpl < 30:
						tpl_state = _("Low")
					
					if tpl < 40 and tpl > 30:
						tpl_state = _("Sub-optimal")

					elif tpl > 40 and rssi < 60:
						tpl_state = _("Optimal")
					
					elif tpl > 60:
						tpl_state = _("High")
					
					elif tpl > 70:
						tpl_state = _("Very High")
						
					if path[1] == self.columns["tpl_pb"]:
						tooltip.set_markup(_("<b>Connected</b>\nReceived Signal Strength: %(rssi)u%% <i>(%(rssi_state)s)</i>\nLink Quality: %(lq)u%%\n<b>Transmit Power Level: %(tpl)u%%</b> <i>(%(tpl_state)s)</i>") % {"rssi_state": rssi_state, "rssi":rssi, "lq":lq, "tpl":tpl, "tpl_state": tpl_state})
					elif path[1] == self.columns["lq_pb"]:
						tooltip.set_markup(_("<b>Connected</b>\nReceived Signal Strength: %(rssi)u%% <i>(%(rssi_state)s)</i>\n<b>Link Quality: %(lq)u%%</b>\nTransmit Power Level: %(tpl)u%% <i>(%(tpl_state)s)</i>") % {"rssi_state": rssi_state, "rssi":rssi, "lq":lq, "tpl":tpl, "tpl_state": tpl_state})
					elif path[1] == self.columns["rssi_pb"]:
						tooltip.set_markup(_("<b>Connected</b>\n<b>Received Signal Strength: %(rssi)u%%</b> <i>(%(rssi_state)s)</i>\nLink Quality: %(lq)u%%\nTransmit Power Level: %(tpl)u%% <i>(%(tpl_state)s)</i>") % {"rssi_state": rssi_state, "rssi":rssi, "lq":lq, "tpl":tpl, "tpl_state": tpl_state})
					
					self.tooltip_row = path[0]
					self.tooltip_col = path[1]
					return True	
		return False
			




# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from gi.repository import GObject
from blueman.ods.OdsBase import OdsBase
from blueman.ods.OdsServerSession import OdsServerSession

class OdsServer(OdsBase):
	__gsignals__ = {
		'started' : (GObject.SignalFlags.NO_HOOKS, None, ()),
		'stopped' : (GObject.SignalFlags.NO_HOOKS, None, ()),
		'closed' : (GObject.SignalFlags.NO_HOOKS, None, ()),
		'error-occured' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)),
		'session-created' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
		'session-removed' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
	}
	
	def __init__(self, obj_path):
		OdsBase.__init__(self, "org.openobex.Server", obj_path)
		
		self.Handle("Started", self.on_started)
		self.Handle("Stopped", self.on_stopped)
		self.Handle("Closed", self.on_closed)
		self.Handle("ErrorOccured", self.on_error)
		self.Handle("SessionCreated", self.on_session_created)
		self.Handle("SessionRemoved", self.on_session_removed)
		
		self.sessions = {}
		
	def __del__(self):
		dprint("deleting server object")
		
	def DisconnectAll(self, *args):
		for k, v in self.sessions.iteritems():
			v.DisconnectAll()
		self.sessions = {}
		OdsBase.DisconnectAll(self, *args)
		
	def on_started(self):
		self.emit("started")
		
	def on_stopped(self):
		self.emit("stopped")
		
	def on_closed(self):
		self.emit("closed")
		self.DisconnectAll()
		
	def on_error(self, err_name, err_message):
		self.emit("error-occured", err_name, err_message)
		self.DisconnectAll()
		
	def on_session_created(self, path):
		dprint(path)
		self.sessions[path] = OdsServerSession(path)
		self.emit("session-created", self.sessions[path])
		
	def on_session_removed(self, path):
		dprint(path)
		self.emit("session-removed", path)
		self.sessions[path].DisconnectAll()
		del self.sessions[path]
	
	
	



import os
from gi.repository import Gtk
from gi.repository import Gdk
import dbus
import gettext
import time
import atexit
import weakref
import pickle
import base64
import zlib
from blueman.Functions import *
from blueman.main.Device import Device
from blueman.bluez.Device import Device as BluezDevice
from blueman.bluez.Adapter import Adapter
from blueman.bluez.Serial import Serial
from blueman.bluez.Network import Network
from blueman.main.SignalTracker import SignalTracker
from blueman.gui.Notification import Notification
import blueman.Sdp as sdp

from blueman.plugins.AppletPlugin import AppletPlugin

REGISTRY_VERSION = 0


class AdapterNotFound(Exception):
    pass


class DeviceNotFound(Exception):
    pass


def store_state():
    try:
        RecentConns.inst.store_state()
    except ReferenceError:
        pass


class RecentConns(AppletPlugin, Gtk.Menu):
    __depends__ = ["Menu"]
    __icon__ = "document-open-recent"
    __description__ = _("Provides a menu item that contains last used connections for quick access")
    __author__ = "Walmis"

    __options__ = {
    "max_items": {"type": int,
                  "default": 6,
                  #the maximum number of items RecentConns menu will display
                  "name": _("Maximum items"),
                  "desc": _("The maximum number of items recent connections menu will display."),
                  "range": (6, 20)
    },
    "recent_connections": {"type": str, "default": ""}
    }

    items = None
    inst = None
    atexit_registered = False

    def on_load(self, applet):
        self.Applet = applet
        self.Adapters = {}
        GObject.GObject.__init__(self)
        if not RecentConns.atexit_registered:
            atexit.register(store_state)
            RecentConns.atexit_registered = True

        self.Item = create_menuitem(_("Recent _Connections") + "...",
                                    get_icon("document-open-recent", 16))

        self.Applet.Plugins.Menu.Register(self, self.Item, 52)
        self.Applet.Plugins.Menu.Register(self, Gtk.SeparatorMenuItem(), 53)

        self.Item.set_submenu(self)

        self.deferred = False
        RecentConns.inst = weakref.proxy(self)

    def store_state(self):
        items = []

        if RecentConns.items:
            for i in RecentConns.items:
                x = i.copy()
                x["device"] = None
                x["mitem"] = None
                x["gsignal"] = 0
                items.append(x)

        try:
            dump = base64.b64encode(
                zlib.compress(
                    pickle.dumps((REGISTRY_VERSION, items),
                                 pickle.HIGHEST_PROTOCOL),
                    9))

            self.set_option("recent_connections", dump)
        except:
            dprint(YELLOW("Failed to store recent connections"))

    def change_sensitivity(self, sensitive):
        try:
            power = self.Applet.Plugins.PowerManager.GetBluetoothStatus()
        except:
            power = True

        sensitive = sensitive and \
                    self.Applet.Manager and \
                    power and \
                    RecentConns.items != None and \
                    (len(RecentConns.items) > 0)

        self.Item.props.sensitive = sensitive

    def on_power_state_changed(self, manager, state):
        self.change_sensitivity(state)
        if state and self.deferred:
            self.deferred = False
            self.on_manager_state_changed(state)


    def on_unload(self):
        self.destroy()
        self.Applet.Plugins.Menu.Unregister(self)
        if RecentConns.items:
            for i in reversed(RecentConns.items):
                if i["device"]:
                    if i["gsignal"]:
                        i["device"].disconnect(i["gsignal"])
                        i["gsignal"] = None

        RecentConns.items = []
        self.destroy()

    def initialize(self):
        dprint("rebuilding menu")
        if not RecentConns.items:
            self.recover_state()

        def compare_by(fieldname):
            def compare_two_dicts(a, b):
                return cmp(a[fieldname], b[fieldname])

            return compare_two_dicts

        def each(child, _):
            self.remove(child)

        self.foreach(each, None)

        RecentConns.items.sort(compare_by("time"), reverse=True)
        for i in RecentConns.items[self.get_option("max_items"):]:
            if i["gsignal"]:
                i["device"].disconnect(i["gsignal"])

        RecentConns.items = RecentConns.items[0:self.get_option("max_items")]
        RecentConns.items.reverse()

        if len(RecentConns.items) == 0:
            self.change_sensitivity(False)
        else:
            self.change_sensitivity(True)

        count = 0
        for item in RecentConns.items:
            if count < self.get_option("max_items"):
                self.add_item(item)
                count += 1

    def on_manager_state_changed(self, state):

        if state:
            try:
                if not self.Applet.Plugins.PowerManager.GetBluetoothStatus():
                    self.deferred = True
                    self.Item.props.sensitive = False
                    return
            except:
                pass

            self.Item.props.sensitive = True
            adapters = self.Applet.Manager.list_adapters()
            self.Adapters = {}
            for adapter in adapters:
                p = adapter.get_properties()
                self.Adapters[str(adapter.get_object_path())] = str(p["Address"])

            if RecentConns.items != None:
                for i in reversed(RecentConns.items):

                    if i["device"]:
                        if i["gsignal"]:
                            i["device"].disconnect(i["gsignal"])
                            i["gsignal"] = 0
                        #i["device"].Destroy()

                    try:
                        i["device"] = self.get_device(i)
                        i["gsignal"] = i["device"].connect("invalidated", self.on_device_removed, i)
                    except:
                        pass

            else:
                self.recover_state()

            self.initialize()
        else:
            self.Item.props.sensitive = False
            return

        self.change_sensitivity(state)

    def on_device_removed(self, device, item):
        device.disconnect(item["gsignal"])
        RecentConns.items.remove(item)
        self.initialize()

    def on_adapter_added(self, path):
        a = Adapter(path)

        def on_activated():
            props = a.get_properties()
            self.Adapters[str(path)] = str(props["Address"])
            self.initialize()

        wait_for_adapter(a, on_activated)

    def on_adapter_removed(self, path):
        try:
            del self.Adapters[str(path)]
        except:
            dprint("Adapter not found in list")

        self.initialize()

    def notify(self, device, service_interface, conn_args):
        dprint(device, service_interface, conn_args)
        item = {}
        object_path = device.get_object_path()
        try:
            adapter = Adapter(device.Adapter)
        except:
            dprint("adapter not found")
            return

        props = adapter.get_properties()

        item["adapter"] = props["Address"]
        item["address"] = device.Address
        item["alias"] = device.Alias
        item["icon"] = device.Icon
        item["service"] = service_interface
        item["conn_args"] = conn_args
        item["time"] = time.time()
        item["device"] = device #device object
        item["mitem"] = None #menu item object

        for i in RecentConns.items:
            if i["adapter"] == item["adapter"] and \
                            i["address"] == item["address"] and \
                            i["service"] == item["service"] and \
                            i["conn_args"] == item["conn_args"]:
                i["time"] = item["time"]
                if i["gsignal"]:
                    i["device"].disconnect(i["gsignal"])

                i["device"] = item["device"]
                i["gsignal"] = item["device"].connect("invalidated", self.on_device_removed, i)
                self.initialize()
                return

        item["gsignal"] = item["device"].connect("invalidated", self.on_device_removed, item)
        RecentConns.items.append(item)
        self.initialize()

        self.store_state()

    def on_item_activated(self, menu_item, item):
        dprint("Connect", item["address"], item["service"])

        sv_name = item["service"].split(".")[-1].lower()
        try:
            service = item["device"].Services[sv_name]
        except:
            RecentConns.items.remove(item)
        else:
            sn = startup_notification("Bluetooth Connection",
                                      desc=_("Connecting to %s") % item["mitem"].get_child().props.label,
                                      icon="blueman")

            item["mitem"].props.sensitive = False

            def reply(*args):
                Notification(_("Connected"), _("Connected to %s") % item["mitem"].get_child().props.label,
                             pixbuf=get_icon("gtk-connect", 48),
                             status_icon=self.Applet.Plugins.StatusIcon)
                item["mitem"].props.sensitive = True
                sn.complete()

            def err(reason):
                Notification(_("Failed to connect"), str(reason).split(": ")[-1],
                             pixbuf=get_icon("gtk-dialog-error", 48),
                             status_icon=self.Applet.Plugins.StatusIcon)
                item["mitem"].props.sensitive = True
                sn.complete()

            if item["service"] == Serial().get_interface_name():
                self.Applet.DbusSvc.RfcommConnect(item["device"].get_object_path(), item["conn_args"][0], reply, err)
            else:
                self.Applet.DbusSvc.ServiceProxy(item["service"], item["device"].get_object_path(), "Connect",
                                                 item["conn_args"], reply, err)

    def add_item(self, item):
        device = item["device"]

        if item["service"] == Serial().get_interface_name():
            name = sdp.sdp_get_serial_name(item["address"], item["conn_args"][0])

        elif item["service"] == Network().get_interface_name():
            name = _("Network Access (%s)") % sdp.uuid16_to_name(sdp.uuid128_to_uuid16(item["conn_args"][0]))
        else:
            try:
                name = sdp.bluez_to_friendly_name(item["service"].split(".")[-1].lower())
            except:
                name = item["service"].split(".")[-1] + " " + _("Service")
                name = name.capitalize()

        if not item["mitem"]:
            mitem = create_menuitem("", get_icon(item["icon"], 16))
            item["mitem"] = mitem
            mitem.connect("activate", self.on_item_activated, item)
        else:
            mitem = item["mitem"]
            mitem.props.sensitive = True
            mitem.props.tooltip_text = None

        item["mitem"].props.label = (_("%(service)s on %(device)s") % {"service": name, "device": item["alias"]})

        if item["adapter"] not in self.Adapters.values():
            if item["device"] and item["gsignal"]:
                item["device"].disconnect(item["gsignal"])

            item["device"] = None
        elif not item["device"] and item["adapter"] in self.Adapters.values():
            try:
                dev = self.get_device(item)
                item["device"] = dev
                item["gsignal"] = item["device"].connect("invalidated", self.on_device_removed, item)

            except:
                RecentConns.items.remove(item)
                self.initialize()

        if not item["device"]:
            mitem.props.sensitive = False
            mitem.props.tooltip_text = _("Adapter for this connection is not available")

        self.prepend(mitem)
        mitem.show()

    def get_device(self, item):
        try:
            adapter = self.Applet.Manager.get_adapter(item["adapter"])
        except:
            raise AdapterNotFound
        try:
            return Device(adapter.find_device(item["address"]))
        except:
            raise DeviceNotFound

    def recover_state(self):
        dump = self.get_option("recent_connections")
        try:
            (version, items) = pickle.loads(zlib.decompress(base64.b64decode(dump)))
        except:
            items = None
            version = None

        if version == None or version != REGISTRY_VERSION:
            items = None

        if items == None:
            RecentConns.items = []
            return

        for i in reversed(items):
            try:
                i["device"] = self.get_device(i)
            except AdapterNotFound:
                i["device"] = None
            except DeviceNotFound:
                items.remove(i)
            else:
                i["gsignal"] = i["device"].connect("invalidated", self.on_device_removed, i)

        RecentConns.items = items



from BlueZInterface import BlueZInterface
from errors import raise_dbus_error
import dbus


class AgentManager(BlueZInterface):
    @raise_dbus_error
    def __init__(self):
        interface = 'org.bluez.AgentManager1'
        super(AgentManager, self).__init__(interface, '/org/bluez')

    @raise_dbus_error
    def register_agent(self, agent, capability='', default=False):
        path = agent.get_object_path()
        self.get_interface().RegisterAgent(path, capability)
        if default:
            self.get_interface().RequestDefaultAgent(path)

    @raise_dbus_error
    def unregister_agent(self, agent):
        self.get_interface().UnregisterAgent(agent.get_object_path())



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import Gtk
from blueman.Constants import *
from blueman.Functions import *

from blueman.gui.GenericList import GenericList
import weakref

class SettingsWidget(Gtk.VBox):
	def __init__(self, inst):
		GObject.GObject.__init__(self)
		self.inst = inst
		self.props.spacing = 2
		
		self.construct_settings()
		self.show_all()
		
	def construct_settings(self):
		for k, v in self.inst.__class__.__options__.iteritems():
			if len(v) > 2:
					
				l = Gtk.Label(label=v["name"])
				l.props.xalign = 0.0
					
				w = self.get_control_widget(k, v)
				if "decorator" in v:
					v["decorator"](self.inst, w, k, v)
				
				self.pack_start(w, False, False, 0)
				
				l = Gtk.Label(label="<i>"+v["desc"]+"</i>")
				l.set_line_wrap(True)
				l.props.use_markup = True
				l.props.xalign = 0.0
				self.pack_start(l, False, False, 0)
				
				sep = Gtk.HSeparator()
				sep.props.height_request = 10
				self.pack_start(sep, False, False, 0)
				
	def handle_change(self, widget, opt, params, prop):
		val = params["type"](getattr(widget.props, prop))
		dprint("changed", opt, val)
		
		self.inst.set_option(opt, val)
		
				
	def get_control_widget(self, opt, params):
		if "widget" in params:
			return params["widget"](self.inst, opt, params)
			
		elif params["type"] == bool:
			c = Gtk.CheckButton(params["name"])
			
			c.props.active = self.inst.get_option(opt)
			c.connect("toggled", self.handle_change, opt, params, "active")
			return c
			
		elif params["type"] == int:
			b = Gtk.HBox()
			l = Gtk.Label(label=params["name"])
			b.pack_start(l, False, False, 0)
			
			r = Gtk.SpinButton()
			b.pack_start(r, False, False, 6)
			b.props.spacing = 6
			
			r.set_numeric(True)
			r.set_increments(1, 3)
			r.set_range(params["range"][0], params["range"][1])
			
			r.props.value = self.inst.get_option(opt)
			r.connect("value-changed", self.handle_change, opt, params, "value")
			
			return b
			
		elif params["type"] == str:
			b = Gtk.HBox()
			l = Gtk.Label(label=params["name"])
			b.pack_start(l, False, False, 0)
			
			e = Gtk.Entry()
			b.pack_start(e, False, False, 6)
			b.props.spacing = 6
			
			e.props.text = self.inst.get_option(opt)
			e.connect("changed", self.handle_change, opt, params, "text")
			
			return b			
				
		
		
		

class PluginDialog(Gtk.Dialog):
	def __init__(self, applet):
		Gtk.Dialog.__init__(self, buttons=(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE))
		
		self.applet = applet
		
		self.Builder = Gtk.Builder()
		self.Builder.set_translation_domain("blueman")
		self.Builder.add_from_file(UI_PATH +"/applet-plugins-widget.ui")
		
		self.set_title(_("Plugins"))
		self.props.icon_name = "blueman"
		
		self.description = self.Builder.get_object("description")
		self.description.props.wrap = True
		
		self.icon = self.Builder.get_object("icon")
		self.author = self.Builder.get_object("author")
		self.depends_hdr = self.Builder.get_object("depends_hdr")
		self.depends_txt = self.Builder.get_object("depends_txt")
		self.conflicts_hdr = self.Builder.get_object("conflicts_hdr")
		self.conflicts_txt = self.Builder.get_object("conflicts_txt")
		self.plugin_name = self.Builder.get_object("name")
		
		self.main_container = self.Builder.get_object("main_container")
		self.content_vbox = self.Builder.get_object("content_vbox")
	
		self.b_prefs = self.Builder.get_object("b_prefs")
		self.b_prefs.connect("toggled", self.on_prefs_toggled)
	
		widget = self.Builder.get_object("all")
		
		ref = weakref.ref(self)
		
		self.vbox.pack_start(widget, True, True, 0)
		
		cr = Gtk.CellRendererToggle()
		cr.connect("toggled", lambda *args: ref() and ref().on_toggled(*args))
		
		data = [
			["active", bool, cr, {"active":0, "activatable":1, "visible":1}, None],
			["activatable", bool],
			["icon", str, Gtk.CellRendererPixbuf(), {"icon-name":2}, None],
			
			#device caption
			["desc", str, Gtk.CellRendererText(), {"markup":3}, None, {"expand": True}],
			["name", str]
		]

		
		self.list = GenericList(data)
		#self.sorted = Gtk.TreeModelSort(self.list.liststore)
		self.list.liststore.set_sort_column_id(3, Gtk.SortType.ASCENDING)
		self.list.liststore.set_sort_func(3, self.list_compare_func)
		
		self.list.selection.connect("changed", lambda *args: ref() and ref().on_selection_changed(*args))
		
		self.list.props.headers_visible = False
		self.list.show()
		
		self.props.border_width = 6
		self.resize(490, 380)
		
		self.Builder.get_object("viewport").add(self.list)
		
		self.populate()
		
		self.sig_a = self.applet.Plugins.connect("plugin-loaded", self.plugin_state_changed, True)
		self.sig_b = self.applet.Plugins.connect("plugin-unloaded", self.plugin_state_changed, False)
		self.connect("response", self.on_response)
		
		self.list.set_cursor(0)
		
	def list_compare_func(self, treemodel, iter1, iter2, user_data):
		a = self.list.get(iter1, "activatable", "name")
		b = self.list.get(iter2, "activatable", "name")

		if (a["activatable"] and b["activatable"]) or (not a["activatable"] and not b["activatable"]):
			if a["name"] == b["name"]:
				return 0
			if a["name"] < b["name"]:
				return -1
			else:
				return 1
		
		else:
			if a["activatable"] and not b["activatable"]:
				return -1
			elif not a["activatable"] and b["activatable"]:
				return 1
		
		
	def on_response(self, dialog, resp):
		self.applet.Plugins.disconnect(self.sig_a)
		self.applet.Plugins.disconnect(self.sig_b)
		
	def on_selection_changed(self, selection):
		model, iter = selection.get_selected()
		
		name = self.list.get(iter, "name")["name"]
		cls = self.applet.Plugins.GetClasses()[name]
		self.plugin_name.props.label = "<b>" + name + "</b>"
		self.icon.props.icon_name = cls.__icon__
		self.author.props.label = cls.__author__ or _("Unspecified")
		self.description.props.label = cls.__description__ or _("Unspecified")
		
		if cls.__depends__ != []:
			self.depends_hdr.props.visible = True
			self.depends_txt.props.visible = True
			self.depends_txt.props.label = ", ".join(cls.__depends__)
		else:
			self.depends_hdr.props.visible = False
			self.depends_txt.props.visible = False
		
		if cls.__conflicts__ != []:
			self.conflicts_hdr.props.visible = True
			self.conflicts_txt.props.visible = True
			self.conflicts_txt.props.label = ", ".join(cls.__conflicts__)
		else:
			self.conflicts_hdr.props.visible = False
			self.conflicts_txt.props.visible = False
			
		if cls.is_configurable() and name in self.applet.Plugins.GetLoaded():
			self.b_prefs.props.sensitive = True
		else:
			self.b_prefs.props.sensitive = False
			
		self.update_config_widget(cls)
		
	def on_prefs_toggled(self, button):
		model, iter = self.list.selection.get_selected()
		name = self.list.get(iter, "name")["name"]
		cls = self.applet.Plugins.GetClasses()[name]
		
		self.update_config_widget(cls)
		
	def update_config_widget(self, cls):
		if self.b_prefs.props.active:
			if not cls.is_configurable():
				self.b_prefs.props.active = False
				return
				
			if not cls.__instance__:
				self.b_prefs.props.active = False
			else:
				c = self.main_container.get_child()
				self.main_container.remove(c)
				if isinstance(c, SettingsWidget):
					c.destroy()
			
				self.main_container.add(SettingsWidget(cls.__instance__))
		
		else:
			c = self.main_container.get_child()
			self.main_container.remove(c)
			if isinstance(c, SettingsWidget):
				c.destroy()
			self.main_container.add(self.content_vbox)
		
	def populate(self):
		classes = self.applet.Plugins.GetClasses()
		loaded = self.applet.Plugins.GetLoaded()
		for name, cls in classes.iteritems():
			if cls.is_configurable():
				desc = "<span weight=\"bold\">%s</span>" % name
			else:
				desc = name
			self.list.append(active=(name in loaded), icon=cls.__icon__, activatable=(cls.__unloadable__), name=name, desc=desc)
			
	def plugin_state_changed(self, plugins, name, loaded):
		row = self.list.get_conditional(name=name)
		self.list.set(row[0], active=loaded)
		
		cls = self.applet.Plugins.GetClasses()[name]
		if not loaded:
			self.update_config_widget(cls)
			self.b_prefs.props.sensitive = False
		elif cls.is_configurable():
			self.b_prefs.props.sensitive = True
			
	def on_toggled(self, cellrenderer, path):
		name = self.list.get(path, "name")["name"]
		
		deps = self.applet.Plugins.GetDependencies()[name]
		loaded = self.applet.Plugins.GetLoaded()
		to_unload = []
		for dep in deps:
			if dep in loaded:
				to_unload.append(dep)
				
		if to_unload != []:
			dialog = Gtk.MessageDialog(self, type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
			dialog.props.secondary_use_markup = True
			dialog.props.icon_name = "blueman"
			dialog.props.text = _("Dependency issue")
			dialog.props.secondary_text = _("Plugin <b>\"%(0)s\"</b> depends on <b>%(1)s</b>. Unloading <b>%(1)s</b> will also unload <b>\"%(0)s\"</b>.\nProceed?") % {"0": ", ".join(to_unload), "1": name}
			
			resp = dialog.run()
			if resp != Gtk.ResponseType.YES:
				dialog.destroy()
				return
			
			dialog.destroy()
		
		conflicts = self.applet.Plugins.GetConflicts()[name]	
		to_unload = []
		for conf in conflicts:
			if conf in loaded:
				to_unload.append(conf)
		
		if to_unload != []:
			dialog = Gtk.MessageDialog(self, type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
			dialog.props.secondary_use_markup = True
			dialog.props.icon_name = "blueman"
			dialog.props.text = _("Dependency issue")
			dialog.props.secondary_text = _("Plugin <b>%(0)s</b> conflicts with <b>%(1)s</b>. Loading <b>%(1)s</b> will unload <b>%(0)s</b>.\nProceed?") % {"0": ", ".join(to_unload), "1": name}
			
			resp = dialog.run()
			if resp != Gtk.ResponseType.YES:
				dialog.destroy()
				return
			
			dialog.destroy()
			
			for p in to_unload:
				self.applet.Plugins.SetConfig(p, False)				
		
		
		loaded = name in self.applet.Plugins.GetLoaded()
		cls = self.applet.Plugins.SetConfig(name, not loaded)
		
		
		
		
		

